from __future__ import annotations
from itertools import chain
import math
import sys
import time
import typing

from tradedangerous.tradedb import describeAge, TradeDB, Station, System
from tradedangerous.tradecalc import NoHopsError, Route, TradeCalc, UserAbortedRun

from .commandenv import ResultRow
from .exceptions import CommandLineError, NoDataError
from .parsing import (
    BlackMarketSwitch, FleetCarrierArgument, MutuallyExclusiveGroup,
    NoPlanetSwitch, OdysseyArgument, PadSizeArgument, ParseArgument,
    PlanetaryArgument,
)

if typing.TYPE_CHECKING:
    from tradedangerous import TradeEnv


######################################################################
# Parser config

help = 'Calculate best trade run.'
name = 'run'
epilog = None
usesTradeData = True

arguments = [
    ParseArgument('--capacity',
            help = 'Maximum capacity of cargo hold.',
            metavar = 'N',
            type = int,
        ),
    ParseArgument('--credits',
            help = 'Starting credits.',
            metavar = 'CR',
            type = "credits",
        ),
]

switches = [
    ParseArgument('--from', '-f',
        help = 'Starting system/station.',
        dest = 'starting',
        metavar = 'STATION',
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--to', '-t',
            help = 'Final system/station.',
            dest = 'ending',
            metavar = 'PLACE',
            default = None,
        ),
        ParseArgument('--towards', '-T',
            help = (
                'Choose a route that continually reduces the '
                'distance towards this system.'
            ),
            dest = 'goalSystem',
            metavar = 'SYSTEM',
            default = None,
        ),
        ParseArgument('--loop',
            help = 'Return to the starting station.',
            action = 'store_true',
            default = False,
        ),
    ),
    ParseArgument('--via',
        help = 'Require specified systems/stations to be en-route.',
        action = 'append',
        metavar = 'PLACE[,PLACE,...]',
    ),
    ParseArgument('--avoid',
        help = 'Exclude an item, system or station from trading. '
                'Partial matches allowed, '
                'e.g. "dom.App" or "domap" matches "Dom. Appliances".',
        action = 'append',
    ),
    MutuallyExclusiveGroup(
        ParseArgument('--direct',
            help = "Assume destinations are reachable without worrying "
                "about jumps.",
            action = 'store_true',
        ),
        ParseArgument('--hops',
            help = 'Number of hops (station-to-station) to run.',
            default = 2,
            type = int,
            metavar = 'N',
        ),
    ),
    ParseArgument('--jumps-per',
        help = 'Maximum number of jumps (system-to-system) per hop.',
        default = 2,
        dest = 'maxJumpsPer',
        metavar = 'N',
        type = int,
    ),
    ParseArgument('--ly-per',
        help = 'Maximum light years per jump.',
        dest = 'maxLyPer',
        metavar = 'N.NN',
        type = float,
    ),
    ParseArgument('--empty-ly',
        help = 'Maximum light years ship can jump when empty.',
        dest = 'emptyLyPer',
        metavar = 'N.NN',
        type = float,
        default = None,
    ),
    ParseArgument('--start-jumps', '-s',
        help = 'Consider stations within this many jumps of the origin '
             '(requires --from).',
        dest = 'startJumps',
        default = 0,
        type = int,
    ),
    ParseArgument('--end-jumps', '-e',
        help = 'Consider stations within this many jumps of the destination '
             '(requires --to).',
        dest = 'endJumps',
        default = 0,
        type = int,
    ),
    ParseArgument('--show-jumps', '-J',
        help = 'Show detail of jumps between hops.',
        dest = 'showJumps',
        action = 'store_true',
    ),
    ParseArgument('--limit',
        help = 'Maximum units of any one cargo item to buy (0: unlimited).',
        metavar = 'N',
        type = int,
    ),
    ParseArgument('--age', '--max-days-old', '-MD',
        help = 'Maximum age (in days) of trade data to use.',
        metavar = 'DAYS',
        type = float,
        dest = 'maxAge',
    ),
    PadSizeArgument(),
    MutuallyExclusiveGroup(
        NoPlanetSwitch(),
        PlanetaryArgument(),
    ),
    FleetCarrierArgument(),
    OdysseyArgument(),
    BlackMarketSwitch(),
    ParseArgument('--ls-penalty', '--lsp',
        help = "Penalty per 1kls stations are from their stars.",
        default = 12.5,
        type = float,
        dest = 'lsPenalty'
    ),
    ParseArgument('--ls-max',
        help = 'Only consider stations upto this many ls from their star.',
        metavar = 'LS',
        dest = 'maxLs',
        type = int,
        default = 0,
    ),
    ParseArgument('--gain-per-ton', '--gpt',
        help = 'Specify the minimum gain per ton of cargo',
        dest = 'minGainPerTon',
        type = "credits",
        default = 1
    ),
    ParseArgument('--max-gain-per-ton', '--mgpt',
        help = 'Specify the maximum gain per ton of cargo',
        dest = 'maxGainPerTon',
        type = "credits",
        default = 0
    ),
    ParseArgument('--unique',
        help = 'Only visit each station once.',
        action = 'store_true',
        default = False,
    ),
    ParseArgument('--loop-interval', '-li',
        help = (
            'Require this many hops between visits to the same station. '
            'A value of 1 would be the default behavior, so a value of '
            '2 is the minimum allowed.'
        ),
        type = int,
        default = None,
        dest = 'loopInt',
    ),
    ParseArgument('--margin',
        help = 'Reduce gains made on each hop to provide a margin of error '
                'for market fluctuations (e.g: 0.25 reduces gains by 1/4). '
                '0<: N<: 0.25.',
        default = 0.00,
        metavar = 'N.NN',
        type = float,
    ),
    ParseArgument('--insurance',
        help = 'Reserve at least this many credits to cover insurance.',
        default = 0,
        metavar = 'CR',
        type = "credits",
    ),
    ParseArgument('--routes',
        help = 'Maximum number of routes to show. DEFAULT: 1',
        default = 1,
        metavar = 'N',
        type = int,
    ),
    ParseArgument('--max-routes',
        help = 'At the end of each hop, limit the number of routes '
                'that continue to the next round to the top N '
                'highest scoring',
        default = 0,
        metavar = 'N',
        type = int,
        dest = 'maxRoutes',
    ),
    ParseArgument('--checklist',
        help = 'Provide a checklist flow for the route.',
        action = 'store_true',
        default = False,
    ),
    ParseArgument('--x52-pro',
        help = 'Enable experimental X52 Pro MFD output.',
        action = 'store_true',
        default = False,
        dest = 'x52pro',
    ),
    ParseArgument('--prune-score',
        help = "From the 3rd hop on, only consider routes which have at least this percentage of the current best route's score.",
        dest = 'pruneScores',
        type = float,
        default = 0,
    ),
    ParseArgument('--prune-hops',
        help = 'Changes which hop --prune-score takes effect from.',
        default = 3,
        type = int,
        dest = 'pruneHops',
    ),
    ParseArgument('--progress', '-P',
        help = 'Show hop progress',
        default = False,
        action = 'store_true',
    ),
    ParseArgument('--supply',
        help = 'Only considers items which have at least this many units.',
        default = None,
        type = int,
    ),
    ParseArgument('--demand',
        help = 'Only considers items which have at least this much demand.',
        default = None,
        type = int
    ),
    ParseArgument('--summary',
        help = 'Summary layout of route instructions.',
        action = 'store_true',
    ),
    ParseArgument('--shorten',
        help = '(Requires --to) Find the shortest route with the best gpt.',
        action = 'store_true',
    ),
]


######################################################################
# Helpers

# Do some basic syntax validation before we waste time loading data.
def validateRunArgumentsFast(cmdenv):
    """
    Fast-fail argument checks that should run BEFORE any database
    access or TradeCalc construction.
    """
    # --towards requires --from
    if cmdenv.goalSystem and not getattr(cmdenv, "starting", None):
        raise CommandLineError("--towards requires --from")
    
    # --start-jumps requires --from
    if cmdenv.startJumps and not getattr(cmdenv, "starting", None):
        raise CommandLineError("--start-jumps requires --from")
    
    # --end-jumps requires --to
    if cmdenv.endJumps and not getattr(cmdenv, "ending", None):
        raise CommandLineError("--end-jumps requires --to")
    
    # --shorten only valid with --to
    if cmdenv.shorten and not cmdenv.destPlace:
        raise CommandLineError("--shorten only works with --to.")

class Checklist:
    """
        Class for encapsulating display of a route as a series of
        steps to be 'checked off' as the user passes through them.
    """
    
    def __init__(self, tdb, cmdenv):
        self.tdb = tdb
        self.cmdenv = cmdenv
        self.mfd = cmdenv.mfd
    
    def doStep(self, action, detail = None, extra = None):
        self.stepNo += 1
        try:
            self.mfd.display(
                "#{} {}".format(self.stepNo, action),
                detail or "",
                extra or ""
            )
        except AttributeError:
            pass
        input(
            "   {:<3}: {}: "
            .format(
                self.stepNo,
                " ".join(item for item in (action, detail, extra) if item)
            )
        )
    
    def note(self, str, addBreak = True):
        print("(i) {} (i){}".format(str, "\n" if addBreak else ""))
    
    def run(self, route, cr):
        mfd = self.mfd
        stations, hops, jumps = route.route, route.hops, route.jumps
        lastHopIdx = len(stations) - 1
        gainCr = 0
        self.stepNo = 0
        
        heading = "(i) BEGINNING CHECKLIST FOR {} (i)".format(route.text(lambda x, y: y))
        print(heading, "\n", '-' * len(heading), "\n\n", sep = '')
        
        cmdenv = self.cmdenv
        if cmdenv.detail:
            print(route.summary())
            print()
        
        for idx in range(lastHopIdx):
            hopNo = idx + 1
            cur, nxt, hop = stations[idx], stations[idx + 1], hops[idx]
            sortedTradeOptions = sorted(
                hop[0],
                key=lambda tradeOption: tradeOption[1] * tradeOption[0].gainCr,
                reverse=True
            )
            
            # Tell them what they need to buy.
            if cmdenv.detail:
                self.note("HOP {} of {}".format(hopNo, lastHopIdx))
            
            self.note("Buy at {}".format(cur.name()))
            for (trade, qty) in sortedTradeOptions:
                self.doStep(
                        'Buy {:n} x'.format(qty),
                        trade.name(),
                        '@ {}cr / {} old'.format(
                            trade.costCr,
                            describeAge(trade.srcAge),
                ))
            if cmdenv.detail:
                self.doStep('Refuel')
            print()
            
            # If there is a next hop, describe how to get there.
            self.note(
                "Fly {}"
                .format(
                    " -> ".join(jump.name() for jump in jumps[idx])
                )
            )
            if idx < len(hops) and jumps[idx]:
                for jump in jumps[idx][1:]:
                    self.doStep('Jump to', jump.name())
            if cmdenv.detail:
                self.doStep('Dock at', nxt.text())
            print()
            
            self.note("Sell at {}".format(nxt.name()))
            for (trade, qty) in sortedTradeOptions:
                self.doStep(
                        'Sell {:n} x'.format(qty),
                        trade.name(),
                        '@ {:n}cr / {} old'.format(
                            trade.costCr + trade.gainCr,
                            describeAge(trade.dstAge),
                ))
            print()
            
            gainCr += hop[1]
            if cmdenv.detail and gainCr > 0:
                self.note("GAINED: {:n}cr, CREDITS: {:n}cr".format(
                            gainCr, cr + gainCr))
            
            if hopNo < lastHopIdx:
                print("\n--------------------------------------\n")
        
        if mfd:
            mfd.display('FINISHED',
                        "+{:n}cr".format(gainCr),
                        "={:n}cr".format(cr + gainCr))
            mfd.attention(3)
            from time import sleep
            sleep(1.5)


def expandForJumps(tdb, cmdenv, calc, origin, jumps, srcName, purpose):
    """
    Find all the stations you could reach if you made a given
    number of jumps away from the origin list.
    """
    
    assert jumps
    
    maxLyPer = cmdenv.emptyLyPer or cmdenv.maxLyPer
    avoidPlaces = cmdenv.avoidPlaces
    cmdenv.DEBUG0(
        "expanding {} reach from {} by {} jumps at {}ly per jump",
        srcName,
        origin.name(),
        jumps,
        maxLyPer,
    )
    
    # Correct behaviour: destinations (--to) require buying data; origins (--from) require selling data.
    if srcName == "--to":
        tradingList = calc.stationsBuying
    elif srcName == "--from":
        tradingList = calc.stationsSelling
    else:
        raise Exception("Unknown src")
    
    # Ensure O(1) membership checks regardless of the underlying container type.
    trading_ids = tradingList if isinstance(tradingList, set) else set(tradingList)
    
    stations: set[Station] = set()
    origins:  set[System | Station] = {origin}
    avoid:    set[System | Station] = set(avoidPlaces)
    
    for jump in range(jumps):
        if not origins:
            break
        if getattr(cmdenv, "debug", False):
            cmdenv.DEBUG1(
                "Ring {}: {}",
                jump,
                [sys.dbname for sys in origins]
            )
        thisJump, origins = origins, set()
        for system in thisJump:
            avoid.add(system)
            for stn in system.stations or ():
                if stn.ID not in trading_ids:
                    if getattr(cmdenv, "debug", False):
                        cmdenv.DEBUG2(
                            "X {}/{} not in trading list",
                            stn.system.dbname, stn.dbname,
                        )
                    continue
                if not checkStationSuitability(cmdenv, calc, stn):
                    if getattr(cmdenv, "debug", False):
                        cmdenv.DEBUG2(
                            "X {}/{} was not suitable",
                            stn.system.dbname, stn.dbname,
                        )
                    continue
                if getattr(cmdenv, "debug", False):
                    cmdenv.DEBUG2(
                        "- {}/{} meets requirements",
                        stn.system.dbname, stn.dbname,
                    )
                stations.add(stn)
            for dest, dist in tdb.genSystemsInRange(system, maxLyPer):
                if dest not in avoid:
                    origins.add(dest)
    
    if getattr(cmdenv, "debug", False):
        cmdenv.DEBUG0(
            "Expanded {} stations: {}",
            srcName,
            [stn.name() for stn in stations]
        )
    
    if not stations:
        if not cmdenv.emptyLyPer:
            extra = (
                "\nIf you are willing to make unladen jumps for the sake "
                "of a better route, consider using --empty."
            )
        else:
            extra = ""
        raise CommandLineError(
            "No {} stations with suitable trade data could be found "
            "within {} {}ly jump{} of {} that meet all of your critera.{}"
            .format(
                purpose, maxLyPer,
                jumps, "s" if jumps > 1 else "",
                origin.name(),
                extra,
            )
        )
    
    stations = list(stations)
    stations.sort(key=lambda stn: stn.ID)
    
    return stations


def checkForEmptyStationList(category, focusPlace, stationList, jumps):
    if stationList:
        return
    if jumps:
        raise NoDataError(
                "Local database has no price data for any "
                "stations within {} jumps of {} ({})".format(
                    jumps,
                    focusPlace.name(),
                    category,
        ))
    if isinstance(focusPlace, System):
        raise NoDataError(
                "Local database either has no price data for "
                "stations in {} ({}) or could not find any that "
                "met your requirements (e.g. pad-size). "
                "Check \"trade.py local -vv --ly 0 {}\"".format(
                    focusPlace.name(),
                    category,
                    focusPlace.name(),
        ))
    raise NoDataError(
            "Local database has no price data for {} ({})".format(
                focusPlace.name(),
                category,
    ))


def checkAnchorNotInVia(hops, anchorName, place, viaSet):
    """
    Ensure that '--to' or '--from' is not in the via set.
    """
    
    if hops != 2:
        return
    if isinstance(place, Station) and place in viaSet:
        raise CommandLineError(
            "{} used in {} and --via with only 2 hops".format(
                place.name(),
                anchorName,
        ))


def checkStationSuitability(cmdenv, calc, station, src = None):
    cmdenv.DEBUG2(
        "checking {} (ls={}, bm={}, pad={}, plt={}, flc={}, ody={}, mkt={}, shp={}) "
        "for {} suitability",
        station.name(),
        station.lsFromStar,
        station.blackMarket,
        station.maxPadSize,
        station.planetary,
        station.fleet,
        station.odyssey,
        station.market,
        station.shipyard,
        src or "any",
    )
    
    if station in cmdenv.avoidPlaces and src != "--from":
        if src:
            raise CommandLineError(
                "{} station {} is marked to avoid"
                .format(src, station.name())
            )
        return False
    if station.system in cmdenv.avoidPlaces and src != "--from":
        if src:
            raise CommandLineError(
                "{} station {} is in system listed in --avoid"
                .format(src, station.name())
            )
        return False
    if station.market == 'N':
        if src:
            raise CommandLineError(
                "{} station {} is flagged as having no market".format(
                    src, station.name()
                )
            )
        return False
    if not station.itemCount:
        if src:
            raise NoDataError(
                "No price data in local database "
                "for {} station: {}".format(
                    src, station.name(),
            ))
        return False
    if src != "--to" and station.ID not in calc.stationsSelling:
        if src:
            raise NoDataError(
                "No buying prices at {}."
                .format(station.name())
            )
        return False
    if src != "--from" and station.ID not in calc.stationsBuying:
        if src:
            raise NoDataError(
                "No selling prices at {}."
                .format(station.name())
            )
        return False
    mps = cmdenv.padSize
    if mps and not station.checkPadSize(mps):
        if src:
            raise CommandLineError(
                "{} station {} does not meet pad-size requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    mps, station.maxPadSize,
                    TradeDB.padSizesExt[station.maxPadSize],
            ))
        return False
    pla = cmdenv.planetary
    if pla and not station.checkPlanetary(pla):
        if src:
            raise CommandLineError(
                "{} station {} does not meet planetary requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    pla, station.planetary,
                    TradeDB.planetStatesExt[station.planetary],
            ))
        return False
    flc = cmdenv.fleet
    if flc and not station.checkFleet(flc):
        if src:
            raise CommandLineError(
                "{} station {} does not meet fleet carrier requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    flc, station.fleet,
                    TradeDB.fleetStatesExt[station.fleet],
            ))
        return False
    ody = cmdenv.odyssey
    if ody and not station.checkOdyssey(ody):
        if src:
            raise CommandLineError(
                "{} station {} does not meet odyssey requirement.\n"
                "You specified: {}, Current data for station: {} ({})\n"
                "You can use \"trade.py station\" to correct this.".format(
                    src, station.name(),
                    ody, station.odyssey,
                    TradeDB.odysseyStatesExt[station.odyssey],
            ))
        return False
    np = cmdenv.noPlanet
    if np and station.planetary != 'N':
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet no-planet "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    bm = cmdenv.blackMarket
    if bm and station.blackMarket != 'Y':
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet black-market "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    mls = cmdenv.maxLs
    if mls and (station.lsFromStar <= 0 or station.lsFromStar > mls):
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet max-ls "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    maxAge, stnAge = cmdenv.maxAge, station.dataAge or float("inf")
    if maxAge and stnAge > maxAge:
        if src and src != "--from":
            raise CommandLineError(
                "{} station {} does not meet --age "
                "requirement.".format(
                    src, station.name(),
            ))
        return False
    return True


def filterStationSet(src, cmdenv, calc, stnList):
    if not stnList:
        return stnList
    if getattr(cmdenv, "debug", False):
        cmdenv.DEBUG0(
            "filtering {} station list: {}",
            src,
            ",".join(station.name() for station in stnList),
        )
    stnList = tuple(
        place for place in stnList
        if isinstance(place, System) or checkStationSuitability(cmdenv, calc, place, src)
    )
    if not stnList:
        # Preserve original message formatting (no behavior change)
        raise CommandLineError("No {src} station met your criteria.")
    return stnList

def checkOrigins(tdb, cmdenv, calc):
    # Compute eligibility once: stations must both sell and buy (same as suitability with src=None).
    eligible_ids = set(calc.stationsSelling) & set(calc.stationsBuying)
    
    setattr(cmdenv, "_origin", cmdenv.origPlace)  # store the original so we can overwrite it
    if cmdenv.origPlace:
        if cmdenv.startJumps and cmdenv.startJumps > 0:
            cmdenv.origins = expandForJumps(
                tdb, cmdenv, calc,
                cmdenv.origPlace.system,
                cmdenv.startJumps,
                "--from", "starting",
            )
            cmdenv.origPlace = None
        elif isinstance(cmdenv.origPlace, System):
            cmdenv.DEBUG0("origPlace: System: {}", cmdenv.origPlace.name())
            if not cmdenv.origPlace.stations:
                raise CommandLineError(
                    "No stations at --from system, {}"
                    .format(cmdenv.origPlace.name())
                )
            cmdenv.origins = tuple(
                station
                for station in cmdenv.origPlace.stations
                if station.ID in eligible_ids and checkStationSuitability(cmdenv, calc, station)
            )
        else:
            cmdenv.DEBUG0("origPlace: Station: {}", cmdenv.origPlace.name())
            checkStationSuitability(cmdenv, calc, cmdenv.origPlace, '--from')
            cmdenv.origins = (cmdenv.origPlace,)
            cmdenv.startStation = cmdenv.origPlace
        checkForEmptyStationList(
            "--from", cmdenv.origPlace,
            cmdenv.origins, cmdenv.startJumps
        )
    else:
        if cmdenv.startJumps:
            raise CommandLineError("--start-jumps (-s) only works with --from")
        cmdenv.DEBUG0("using all suitable origins")
        cmdenv.origins = tuple(
            station
            for station in tdb.stationByID.values()
            if station.ID in eligible_ids and checkStationSuitability(cmdenv, calc, station)
        )
    
    if not cmdenv.startJumps and isinstance(cmdenv.origPlace, System):
        cmdenv.origins = filterStationSet(
            '--from', cmdenv, calc, cmdenv.origins
        )
    
    cmdenv.origSystems = tuple(set(
        stn.system for stn in cmdenv.origins
    ))

def checkDestinations(tdb, cmdenv, calc):
    cmdenv.destinations = None
    showProgress = bool(getattr(cmdenv, "progress", False))
    hb_interval = 0.5
    last_hb = 0.0
    spinner = ("|", "/", "-", "\\")
    spin_i = 0
    
    def heartbeat(seen, kept):
        nonlocal last_hb, spin_i
        if not showProgress:
            return
        now = time.time()
        if (now - last_hb) < hb_interval:
            return
        last_hb = now
        s = spinner[spin_i]
        spin_i = (spin_i + 1) % len(spinner)
        sys.stdout.write(
            f"\r{s} Scanning stations…  examined {seen:n}  kept {kept:n}"
        )
        sys.stdout.flush()
    
    if cmdenv.destPlace:
        if cmdenv.endJumps and cmdenv.endJumps > 0:
            cmdenv.destinations = expandForJumps(
                    tdb, cmdenv, calc,
                    cmdenv.destPlace.system,
                    cmdenv.endJumps,
                    "--to", "destination",
            )
            cmdenv.destPlace = None
        elif isinstance(cmdenv.destPlace, Station):
            cmdenv.DEBUG0("destPlace: Station: {}", cmdenv.destPlace.name())
            # Single-station --to: hard fail if unsuitable.
            checkStationSuitability(cmdenv, calc, cmdenv.destPlace, '--to')
            cmdenv.destinations = (cmdenv.destPlace,)
        else:
            cmdenv.DEBUG0("destPlace: System: {}", cmdenv.destPlace.name())
            # System --to: examine all stations, keeping only those that pass --to suitability.
            dests, seen, kept = [], 0, 0
            for station in cmdenv.destPlace.stations:
                seen += 1
                try:
                    ok = checkStationSuitability(cmdenv, calc, station, '--to')
                except (CommandLineError, NoDataError):
                    ok = False
                if ok:
                    dests.append(station)
                    kept += 1
                heartbeat(seen, kept)
            cmdenv.destinations = tuple(dests)
            if showProgress:
                sys.stdout.write("\n")
                sys.stdout.flush()
        checkForEmptyStationList(
                "--to", cmdenv.destPlace,
                cmdenv.destinations, cmdenv.endJumps
        )
    else:
        if cmdenv.endJumps:
            raise CommandLineError("--end-jumps (-e) only works with --to")
        cmdenv.DEBUG0("Using all available destinations")
        if cmdenv.goalSystem:
            dest = tdb.lookupPlace(cmdenv.goalSystem)
            cmdenv.goalSystem = dest.system
        
        if cmdenv.origPlace and cmdenv.maxJumpsPer == 0:
            stationSrc = chain.from_iterable(
                system.stations for system in cmdenv.origSystems
            )
        else:
            stationSrc = tdb.stationByID.values()
        
        eligible_ids = set(calc.stationsSelling) & set(calc.stationsBuying)
        
        dests, seen, kept = [], 0, 0
        for station in stationSrc:
            seen += 1
            if station.ID in eligible_ids and checkStationSuitability(cmdenv, calc, station):
                dests.append(station)
                kept += 1
            heartbeat(seen, kept)
        cmdenv.destinations = tuple(dests)
        if showProgress:
            sys.stdout.write("\n")
            sys.stdout.flush()
    
    if not cmdenv.endJumps and isinstance(cmdenv.destPlace, System):
        cmdenv.destinations = filterStationSet(
            '--to', cmdenv, calc, cmdenv.destinations
        )
    
    cmdenv.destSystems = tuple(set(
        stn.system for stn in cmdenv.destinations
    ))


def validateRunArguments(tdb, cmdenv, calc):
    """
        Process arguments to the 'run' option.
    """
    
    if cmdenv.credits < 0:
        raise CommandLineError("Invalid (negative) value for initial credits")
    # I'm going to allow 0 credits as a future way of saying "just fly"
    
    if cmdenv.routes < 1:
        raise CommandLineError(
            "Maximum routes has to be 1 or higher."
        )
    if cmdenv.routes > 1 and cmdenv.checklist:
        raise CommandLineError(
            "Checklist can only be applied to a single route."
        )
    
    if cmdenv.hops < 1:
        raise CommandLineError("Minimum of 1 hop required")
    if cmdenv.hops > 32:
        raise CommandLineError("Too many hops without more optimization")
    
    if cmdenv.maxJumpsPer < 0:
        raise CommandLineError("Negative jumps: you're already there?")
    if cmdenv.direct:
        cmdenv.hops = 1
        cmdenv.maxJumpsPer = cmdenv.maxLyPer = 10000
    
    if cmdenv.capacity is None:
        raise CommandLineError("Missing '--capacity'")
    if cmdenv.maxLyPer is None and not cmdenv.direct:
        raise CommandLineError("Missing '--ly-per'")
    if cmdenv.capacity < 0:
        raise CommandLineError("Invalid (negative) cargo capacity")
    if cmdenv.capacity > 1500:
        cmdenv.WARN("Capacity > 1500 not supported (you specified {})", cmdenv.capacity)
        cmdenv.WARN("Forcing jumps per hop to 1.")
        cmdenv.maxJumpsPer = 1
        if cmdenv.hops > 2:
            cmdenv.WARN("{} hops? Press [CTRL][C] to quit.", cmdenv.hops)
        if not cmdenv.supply:
            cmdenv.WARN("Please provide a '--supply' value.")
            cmdenv.supply = cmdenv.capacity * 10
            cmdenv.DEBUG0("'supply' minimum set to {}.", cmdenv.supply)
        if not cmdenv.demand:
            cmdenv.WARN("Please provide a '--demand' value.")
            cmdenv.demand = cmdenv.capacity * 10
            cmdenv.DEBUG0("'demand' minimum set to {}.", cmdenv.demand)
    #    raise CommandLineError(
    #        "Capacity > 1500 not supported (you specified {})"
    #        .format(cmdenv.capacity)
    #    )
    
    if cmdenv.limit and cmdenv.limit > cmdenv.capacity:
        raise CommandLineError("'limit' must be <= capacity")
    if cmdenv.limit and cmdenv.limit < 0:
        raise CommandLineError("'limit' can't be negative, silly")
    cmdenv.maxUnits = cmdenv.limit if cmdenv.limit else cmdenv.capacity
    
    if cmdenv.insurance:
        arbitraryInsuranceBuffer = 42
        if cmdenv.insurance >= (cmdenv.credits + arbitraryInsuranceBuffer):
            raise CommandLineError("Insurance leaves no margin for trade")
    
    if cmdenv.loop:
        if cmdenv.unique:
            raise CommandLineError("Cannot use --unique and --loop together")
        if cmdenv.direct:
            raise CommandLineError("Cannot use --direct and --loop together")
    
    if cmdenv.loopInt:
        if cmdenv.loopInt < 2:
            raise CommandLineError(
                "--loop-int must be 2 or higher to have any effect. "
            )
        if cmdenv.loopInt > cmdenv.hops and not cmdenv.unique:
            cmdenv.NOTE("--loop-int > hops implies --unique")
            cmdenv.unique = True
    
    if cmdenv.shorten:
        if cmdenv.loop:
            raise CommandLineError(
                "Cannot use --shorten and --loop together"
            )
        if not cmdenv.ending:
            raise CommandLineError(
                "--shorten only works with --to."
            )
    
    if cmdenv.goalSystem and not cmdenv.origPlace:
        raise CommandLineError("--towards requires --from")
    
    checkOrigins(tdb, cmdenv, calc)
    checkDestinations(tdb, cmdenv, calc)
    
    # If they're going --from and --to single systems, and they have
    # specified zero jumps then it's futile to try anything.
    if cmdenv.maxJumpsPer == 0 and not cmdenv.direct:
        if len(cmdenv.origSystems) == 1 and len(cmdenv.destSystems) == 1:
            if cmdenv.origSystems[0] != cmdenv.destSystems[0]:
                raise CommandLineError(
                    "Could not find any connections that didn't require at "
                    "least one jump and --jumps 0 specified."
                )
    
    origins, destns = cmdenv.origins or (), cmdenv.destinations or ()
    
    if cmdenv.hops == 1 and len(origins) == 1 and len(destns) == 1:
        if origins == destns:
            raise CommandLineError("Same to/from; more than one hop required.")
    
    avoidSet = set(cmdenv.avoidPlaces or ())
    viaSet = cmdenv.viaSet = set(cmdenv.viaPlaces)
    cmdenv.DEBUG0("Via: {}", viaSet)
    cmdenv.viaSet = filterStationSet('--via', cmdenv, calc, cmdenv.viaSet)
    checkAnchorNotInVia(cmdenv.hops, "--from", cmdenv.origPlace, viaSet)
    checkAnchorNotInVia(cmdenv.hops, "--to", cmdenv.destPlace, viaSet)
    
    viaSystems = set()
    for place in viaSet:
        if place in avoidSet or place.system in avoidSet:
            raise CommandLineError(
                '"--via {}" conflicts with --avoid'
                .format(place.name())
            )
        if isinstance(place, Station):
            viaSystems.add(place.system)
        else:
            viaSystems.add(place)
    
    if cmdenv.maxJumpsPer == 0 and viaSet and not cmdenv.direct:
        for via in viaSet:
            if via.system not in cmdenv.origSystems:
                raise CommandLineError(
                    "--via {} unreachable with --jumps 0"
                    .format(via.name())
                )
        cmdenv.origins = tuple(
            origin for origin in cmdenv.origins
            if origin.system in viaSystems
        )
        cmdenv.origSystems = tuple(
            origin.system for origin in cmdenv.origins
        )
        cmdenv.destinations = tuple(
            dest for dest in cmdenv.destinations
            if dest.system in viaSystems
        )
        cmdenv.destSystems = tuple(
            dest.system for dest in cmdenv.destinations
        )
    
    # How many of the hops do not have pre-determined stations. For example,
    # when the user uses "--from", they pre-determine the starting station.
    fixedRoutePoints = 0
    if cmdenv.origPlace:
        fixedRoutePoints += 1
    if cmdenv.destPlace:
        fixedRoutePoints += 1
    totalRoutePoints = cmdenv.hops + 1
    adhocRoutePoints = totalRoutePoints - fixedRoutePoints
    if len(viaSystems) > adhocRoutePoints:
        raise CommandLineError(
                "Route is not long enough for the list of '--via' "
                "destinations you gave. Reduce the vias or try again "
                "with '--hops {}' or greater.\n".format(
                    len(viaSet) + fixedRoutePoints - 1
                ))
    cmdenv.adhocHops = adhocRoutePoints - 1
    
    if cmdenv.unique and cmdenv.hops >= len(tdb.stationByID):
        raise CommandLineError(
            "Requested unique trip with more hops than there are stations..."
        )
    if cmdenv.unique:
        # if there's only one start and stop...
        if len(origins) == 1 and len(destns) == 1:
            if origins[0] is destns[0]:
                raise CommandLineError("Can't have same from/to with --unique")
        if viaSet:
            if len(origins) == 1 and origins[0] in viaSet:
                raise CommandLineError("Can't have --from station in --via list with --unique")
            if len(destns) == 1 and destns[0] in viaSet:
                raise CommandLineError("Can't have --to station in --via list with --unique")
    
    if cmdenv.mfd:
        cmdenv.mfd.display("Loading Trades")
    
    if cmdenv.pruneScores and cmdenv.pruneHops:
        if cmdenv.pruneScores > 99:
            raise CommandLineError("--prune-score value percentile exceed 99.")
        if cmdenv.pruneHops < 2:
            raise CommandLineError("--prune-hops must 2 or more.")
    else:
        cmdenv.pruneScores = cmdenv.pruneHops = 0

######################################################################


def filterByVia(routes, viaSet, viaStartPos):
    if not routes:
        return ()
    
    matchedRoutes = []
    partialRoutes = {}
    maxMet = 0
    for route in routes:
        met = 0
        for hop in route.route[viaStartPos:]:
            if hop in viaSet or hop.system in viaSet:
                met += 1
        if met > 0:
            if met >= len(viaSet):
                matchedRoutes.append(route)
            else:
                if met > maxMet:
                    partialRoutes[met] = []
                if met >= maxMet:
                    maxMet = met
                    partialRoutes[met].append(route)
    
    if matchedRoutes:
        return matchedRoutes, None
    
    if not maxMet:
        raise NoDataError(
                "No routes were found which matched your 'via' selections."
        )
    
    return partialRoutes[maxMet], (
            "SORRY: No runs visited all of your via destinations. "
            "Listing runs that matched at least {}.".format(
                    maxMet
            )
    )


def checkReachability(tdb, cmdenv):
    if cmdenv.direct:
        return
    srcSys, dstSys = cmdenv.origSystems, cmdenv.destSystems
    if len(srcSys) == 1 and len(dstSys) == 1:
        srcSys, dstSys = srcSys[0], dstSys[0]
        if srcSys != dstSys:
            maxLyPer = cmdenv.maxLyPer
            avoiding = tuple(
                avoid for avoid in cmdenv.avoidPlaces
                if isinstance(avoid, System)
            )
            route = tdb.getRoute(
                srcSys, dstSys, maxLyPer, avoiding,
            )
            if not route:
                raise CommandLineError(
                    "No route between {} and {} with a {}ly/jump limit."
                    .format(
                        srcSys.name(), dstSys.name(),
                        maxLyPer,
                    )
                )
            
            # Were there just not enough hops?
            jumpLimit = cmdenv.maxJumpsPer * cmdenv.hops
            routeJumps = len(route) - 1
            if jumpLimit < routeJumps:
                hopsRequired = math.ceil(routeJumps / cmdenv.maxJumpsPer)
                jumpsRequired = math.ceil(routeJumps / cmdenv.hops)
                raise CommandLineError(
                    "Shortest route between {src} and {dst} at {jumply} "
                    "ly per jump requires at least {minjumps} jumps. "
                    "Your current settings (--hops {hops} --jumps {jumps}) "
                    "allows a maximum of {jumplimit}.\n"
                    "\n"
                    "You may need --hops={althops} or --jumps={altjumps}.\n"
                    "\n"
                    "See also:\n"
                    " --towards (aka -T),"
                    " --start-jumps (-s),"
                    " --end-jumps (-e),"
                    " --direct.\n"
                    .format(
                        src = srcSys.name(),
                        dst = dstSys.name(),
                        jumply = cmdenv.maxLyPer,
                        minjumps = routeJumps,
                        hops = cmdenv.hops,
                        jumps = cmdenv.maxJumpsPer,
                        jumplimit = jumpLimit,
                        althops = hopsRequired,
                        altjumps = jumpsRequired,
                    )
                )


def routeFailedRestrictions(
        tdb, cmdenv, restrictTo, maxLs, hopNo
        ):
    """
    Generate exception text indicating we couldn't complete a
    route given the restrictions supplied. If the user has
    specified detail, check if there is a route at all.
    """
    
    places = list(
        set(
            chain.from_iterable(
                (place,) if isinstance(place, Station) else place.stations
                for place in restrictTo
            )
        )
    )
    places.sort(key = lambda stn: stn.dbname)
    
    dests = ", ".join(place.name() for place in places)
    
    return (
        "SORRY: Could not find any routes that delivered a profit to "
        "{} at hop #{}\n"
        "You may need to add more hops to your route or adjust your "
        "filters/restrictions.\n"
        .format(
            dests, hopNo + 1
        )
    )


def extraRouteProgress(routes):
    bestGain = max(routes, key = lambda route: route.gainCr).gainCr
    worstGain = min(routes, key = lambda route: route.gainCr).gainCr
    if bestGain != worstGain:
        gainText = "{:n}-{:n}cr gain".format(worstGain, bestGain)
    else:
        gainText = "{:n}cr gain".format(bestGain)
    
    bestGPT = int(max(routes, key = lambda route: route.gpt).gpt)
    worstGPT = int(min(routes, key = lambda route: route.gpt).gpt)
    if bestGPT != worstGPT:
        gptText = "{:n}-{:n}cr/ton".format(worstGPT, bestGPT)
    else:
        gptText = "{:n}cr/ton".format(bestGPT)
    
    return ".. {}, {}".format(gainText, gptText)

######################################################################
# Perform query and populate result set


def run(results, cmdenv, tdb):
    cmdenv.DEBUG1("loading trades")
    
    if tdb.tradingCount == 0:
        raise NoDataError("Database does not contain any profitable trades.")
    
    # Always show a friendly heads-up before heavy work begins.
    print("Searching for quality trades. This may take a few minutes. Please be patient.", flush=True)
    
    # Instantiate the calculator object
    calc = TradeCalc(tdb, cmdenv)
    
    validateRunArguments(tdb, cmdenv, calc)
    
    origPlace, viaSet = cmdenv.origPlace, cmdenv.viaSet
    stopStations = cmdenv.destinations
    goalSystem = cmdenv.goalSystem
    maxLs = cmdenv.maxLs
    
    # seed the route table with starting places
    startCr = cmdenv.credits - cmdenv.insurance
    routes = [
        Route(
            stations = (src,),
            hops = (),
            jumps = (),
            startCr = startCr,
            gainCr = 0,
            score = 0,
        )
        for src in cmdenv.origins
    ]
    
    numHops = cmdenv.hops
    lastHop = numHops - 1
    viaStartPos = 1 if origPlace else 0
    
    cmdenv.DEBUG1("numHops {}, vias {}, adhocHops {}",
                numHops, len(viaSet), cmdenv.adhocHops)
    
    results.summary = ResultRow()
    results.summary.exception = ""
    
    if cmdenv.loop:
        routePickPred = lambda route: \
            route.lastStation is route.firstStation
    elif cmdenv.shorten:
        if not cmdenv.destPlace:
            routePickPred = lambda route: \
                route.lastStation is route.firstStation
        elif isinstance(cmdenv.destPlace, System):
            routePickPred = lambda route: \
                route.lastSystem is cmdenv.destPlace
        else:
            routePickPred = lambda route: \
                route.lastStation is cmdenv.destPlace
    else:
        routePickPred = None
    
    pickedRoutes = []
    
    pruneMod = cmdenv.pruneScores / 100
    
    if cmdenv.loop:
        distancePruning = lambda rt, distLeft: \
            rt.lastSystem.distanceTo(rt.firstSystem) <= distLeft
    elif cmdenv.destPlace and not cmdenv.direct:
        distancePruning = lambda rt, distLeft: \
            any(
                stop for stop in stopSystems
                if rt.lastSystem.distanceTo(stop) <= distLeft
            )
    else:
        distancePruning = False
    
    if distancePruning:
        maxHopDistLy = cmdenv.maxJumpsPer * cmdenv.maxLyPer
        if not cmdenv.loop:
            stopSystems = {stop.system for stop in stopStations}
    
    for hopNo in range(numHops):
        restrictTo = None
        if hopNo == lastHop and stopStations:
            restrictTo = set(stopStations)
            manualRestriction = bool(cmdenv.destPlace)
        elif len(viaSet) > cmdenv.adhocHops:
            restrictTo = viaSet
            manualRestriction = True
        
        if distancePruning:
            preCrop = len(routes)
            distLeft = maxHopDistLy * (numHops - hopNo)
            routes[:] = [rt for rt in routes if distancePruning(rt, distLeft)]
            if not routes:
                if pickedRoutes:
                    break
                raise NoDataError(
                    "No routes are in-range of any end stations at the end of hop {}"
                    .format(hopNo)
                )
            if (pruned := preCrop - len(routes)):
                cmdenv.NOTE("Pruned {} origins too far from any end stations", pruned)
        
        if hopNo >= 1 and (cmdenv.maxRoutes or pruneMod):
            routes.sort()
            if pruneMod and hopNo + 1 >= cmdenv.pruneHops and len(routes) > 10:
                crop = int(len(routes) * pruneMod)
                routes[:] = routes[:-crop]
                cmdenv.NOTE("Pruned {} origins", crop)
            
            if cmdenv.maxRoutes and len(routes) > cmdenv.maxRoutes:
                routes[:] = routes[:cmdenv.maxRoutes]
        
        if cmdenv.progress:
            extra = ""
            if hopNo > 0 and cmdenv.detail > 1:
                extra = extraRouteProgress(routes)
            print(
                "* Hop {:3n}: {:.>10n} origins {}"
                .format(hopNo + 1, len(routes), extra)
            )
        elif cmdenv.debug:
            cmdenv.DEBUG0("Hop {}...", hopNo + 1)
        
        try:
            newRoutes = calc.getBestHops(routes, restrictTo = restrictTo)
        
        except KeyboardInterrupt:
            cmdenv.DEBUG0("** Keyboard Interrupt")
            if hopNo == 0 or not routes:
                raise UserAbortedRun("before any routes calculated")
            # Until python 3.14 it's discouraged to break from an exception, so
            # lets make sure we don't mistake there being anything to process.
            calc.aborted = True
            newRoutes = []
        
        except NoHopsError:
            if hopNo == 0 and len(cmdenv.origSystems) == 1:
                raise NoDataError(
                    "Couldn't find any trading links within {} x {}ly jumps of {}."
                    .format(
                        cmdenv.maxJumpsPer,
                        cmdenv.maxLyPer,
                        cmdenv.origSystems[0].name(),
                    )
                )
            raise NoDataError(
                "No routes had reachable trading links at hop #{}".format(hopNo + 1)
            )
        
        if calc.aborted:
            cmdenv.DEBUG0("** User Aborted")
            break
        
        if not newRoutes:
            assert not calc.aborted, "internal error"
            # First attempt to find a route is a special case because the current
            # route list is the source.
            if hopNo == 0:
                no_routes_on_first_hop(cmdenv, calc)
                # no return
            
            # If we've already got some winners (e.g. on --shorten)
            if pickedRoutes:
                break
            
            checkReachability(tdb, cmdenv)
            
            if restrictTo and manualRestriction:
                results.summary.exception += routeFailedRestrictions(
                    tdb, cmdenv, restrictTo, maxLs, hopNo
                )
                break
            
            results.summary.exception += f"SORRY: Could not find profitable destinations beyond hop #{hopNo+1:n}\n"
            break
        
        routes[:] = newRoutes
        if goalSystem:
            # Promote the winning route to the top of the list
            # while leaving the remainder of the list intact
            routes.sort(
                key = lambda route:
                    0 if route.lastSystem is goalSystem else 1
            )
            if routes[0].lastSystem is goalSystem:
                cmdenv.NOTE("Goal system reached!")
                routes = routes[:1]
                break
        
        if calc.aborted:
            break
        
        if routePickPred:
            pickedRoutes.extend(
                route for route in routes if routePickPred(route)
            )
    
    if cmdenv.loop or cmdenv.shorten:
        cmdenv.DEBUG0("Using {} picked routes", len(pickedRoutes))
        routes = pickedRoutes
        # normalise the scores for fairness...
        for route in routes:
            cmdenv.DEBUG0(
                "{} hops, {} score, {} gpt",
                len(route.hops), route.score, route.gpt
            )
            route.score /= len(route.hops)
    
    if not routes:
        if calc.aborted:
            raise UserAbortedRun("before any routes found")
        raise NoDataError(
            "No profitable trades matched your critera, or price data along the route is missing."
        )
    
    if viaSet:
        routes, caution = filterByVia(routes, viaSet, viaStartPos)
        if caution:
            results.summary.exception += caution + "\n"
    
    routes.sort()
    results.data = routes
    
    if calc.aborted:
        results.summary.exception += str(UserAbortedRun("results may be incomplete or inaccurate")) + "\n"
    
    return results


def no_routes_on_first_hop(cmdenv: TradeEnv, calc: TradeCalc) -> None:
    """ handle the special case where run found no routes on the first hop. """
    # Is it because you ctrl-c'd?
    if calc.aborted:
        raise UserAbortedRun("during first hop before any routes found")
    
    # The raw name they provide with --from is stored as cmdenv.starting, and resolved
    # to a System or Station in cmdenv.origPlace, however checkOrigins may set that to
    # None if we're doing --start-jumps to indicate there's no "single" origin. So we
    # saved a copy of it to cmdenv._origin.
    start_place = getattr(cmdenv, "_origin")
    if not start_place:
        # Ok, we were doing some kind of open-ended galaxy wide query
        raise NoDataError("Could not find any trade links in the galaxy with those criteria.")
    
    # Find the system name - all "locations" have a system property including Systems.
    start_system = start_place.system.name()
    
    # How far did you say you were willing to go?
    max_ly = cmdenv.maxJumpsPer * cmdenv.maxLyPer
    
    errText = (
        f"No suitable and profitable buyers found at/relative to {start_place}.\n"
        "\n"
        "You may want to try:\n"
        f"  {sys.argv[0]} local \"{start_system}\" --ly {max_ly} -vv --stations --trading"
    )
    
    # If they had specified a station, give them a little extra help.
    if isinstance(start_place, Station):
        errText += (
            "\n"
            "or:\n"
            f"  {sys.argv[0]} market \"{start_place}\" --sell -vv"
        )
    
    raise NoDataError(errText)


######################################################################
# Transform result set into output


def render(results, cmdenv, tdb):
    if (exception := results.summary.exception.strip()):
        style = ""
        lines = exception.split("\n")
        max_line_len = max(len(line) for line in lines)
        if cmdenv.color:
            style = "yellow on grey15"  # yellow on a darkish background, so we're sure it's not on a light background
            # Pad all the lines to the same length
            exception = "\n".join(f"{line:{max_line_len}s}" for line in lines)
        
        # TODO: should use a rich panel when --color is set
        cmdenv.console.print('#' * max_line_len, style=style)
        cmdenv.console.print(exception, style=style)
        cmdenv.console.print('#' * max_line_len, style=style)
        # Ring the console bell and add a blank line
        cmdenv.console.print("\a")
    
    routes = results.data
    
    for i in range(min(len(routes), cmdenv.routes)):
        print(routes[i].detail(cmdenv))
    
    # User wants to be guided through the route.
    if cmdenv.checklist:
        assert cmdenv.routes == 1
        cl = Checklist(tdb, cmdenv)
        cl.run(routes[0], cmdenv.credits)



import pytest
from collections import namedtuple

from tradedangerous import cache

FakeFile = namedtuple('FakeFile', ['name'])

class TestCache:
    def test_parseSupply(self):
        fil = FakeFile('faked-file.prices')
        reading = '897H'
        demandUnits, demandLevel = cache.parseSupply(
            fil,
            10,
            'demand',
            reading)
        
        assert demandUnits == 897
        assert demandLevel == 3
    
    def test_parseSupply_bad_level(self):
        fil = FakeFile('faked-file.prices')
        reading = '897X'
        with pytest.raises(cache.SupplyError, match='Unrecognized level suffix'):
            cache.parseSupply(
                fil,
                10,
                'demand',
                reading)
    
    def test_parseSupply_bad_units(self):
        fil = FakeFile('faked-file.prices')
        reading = '-10H'
        with pytest.raises(cache.SupplyError, match='Unrecognized units/level value'):
            cache.parseSupply(
                fil,
                10,
                'demand',
                reading)



"""
Import plugin that uses data files from
https://elite.tromador.com/ to update the Database.
"""
from __future__ import annotations

from contextlib import contextmanager
from email.utils import parsedate_to_datetime
from pathlib import Path
import csv
import datetime
import os
import requests
import time
import typing

from sqlalchemy.orm import Session
from sqlalchemy import func, delete, select, exists, text

from tradedangerous import plugins, transfers, TradeException
from tradedangerous.db import orm_models as SA, lifecycle
from tradedangerous.db.utils import (
    begin_bulk_mode, end_bulk_mode,
    get_import_batch_size, get_upsert_fn,
)
from tradedangerous.fs import file_line_count
from tradedangerous.misc import progress as pbar
from tradedangerous.plugins import PluginException

if typing.TYPE_CHECKING:
    from tradedangerous.tradeenv import TradeEnv


# Constants
BASE_URL = os.environ.get('TD_SERVER') or "https://elite.tromador.com/files/"


class DecodingError(PluginException):
    pass


@contextmanager
def bench(label: str, tdenv: TradeEnv):
    started = time.time()
    with pbar.Progress(0, 40, label=label, style=pbar.ElapsedBar):
        yield
    tdenv.NOTE("{} done ({:.3f}s)", label, time.time() - started)


def _count_listing_entries(tdenv: TradeEnv, listings: Path) -> int:
    """ Calculates the number of entries in a listing file by counting the lines. """
    if not listings.exists():
        tdenv.NOTE("File not found, aborting: {}", listings)
        return 0
    
    tdenv.DEBUG0(f"Getting total number of entries in {listings}...")
    count = file_line_count(listings)
    if count <= 1:
        if count == 1:
            tdenv.DEBUG0("Listing count of 1 suggests nothing but a header")
        else:
            tdenv.DEBUG0("Listings file is empty, nothing to do.")
        return 0
    
    return count + 1  # kfsone: Doesn't the header already make this + 1?


def _make_item_id_lookup(tdenv: TradeEnv, session: Session) -> frozenset[int]:
    """Helper: retrieve the list of commodities in database."""
    tdenv.DEBUG0("Getting list of commodities...")
    rows = session.query(SA.Item.item_id).all()
    return frozenset(r[0] for r in rows)


def _make_station_id_lookup(tdenv: TradeEnv, session: Session) -> frozenset[int]:
    """Helper: retrieve the list of station IDs in database."""
    tdenv.DEBUG0("Getting list of stations...")
    rows = session.query(SA.Station.station_id).all()
    return frozenset(r[0] for r in rows)


def _collect_station_modified_times(tdenv: TradeEnv, session: Session) -> dict[int, int]:
    """Helper: build a list of the last modified time for all stations by id (epoch seconds)."""
    tdenv.DEBUG0("Getting last-update times for stations...")
    rows = (
        session.query(
            SA.StationItem.station_id,
            func.min(SA.StationItem.modified),
        )
        .group_by(SA.StationItem.station_id)
        .all()
    )
    return {
        station_id: int(modified.timestamp()) if modified else 0
        for station_id, modified in rows
    }


class ImportPlugin(plugins.ImportPluginBase):
    """
    Import plugin that uses data files from
    https://elite.tromador.com/ to update the Database.
    """
    pluginOptions = {
        'item':         "Update Items using latest file from server. (Implies '-O system,station')",
        'rare':         "Update RareItems using latest file from server. (Implies '-O system,station')",
        'ship':         "Update Ships using latest file from server.",
        'upgrade':      "Update Upgrades using latest file from server.",
        'system':       "Update Systems using latest file from server.",
        'station':      "Update Stations using latest file from server. (Implies '-O system')",
        'shipvend':     "Update ShipVendors using latest file from server. (Implies '-O system,station,ship')",
        'upvend':       "Update UpgradeVendors using latest file from server. (Implies '-O system,station,upgrade')",
        'listings':     "Update market data using latest listings.csv dump. (Implies '-O item,system,station')",
        'all':          "Update everything with latest dumpfiles. (Regenerates all tables)",
        'clean':        "Erase entire database and rebuild from empty. (Regenerates all tables.)",
        'skipvend':     "Don't regenerate ShipVendors or UpgradeVendors. (Supercedes '-O all', '-O clean'.)",
        'force':        "Force regeneration of selected items even if source file not updated since previous run. "
                        "(Useful for updating Vendor tables if they were skipped during a '-O clean' run.)",
        'purge':        "Remove any empty systems that previously had fleet carriers.",
        'optimize':     "Optimize ('vacuum') database after processing.",
        'solo':         "Don't download crowd-sourced market data. "
                        "(Implies '-O skipvend', supercedes '-O all', '-O clean', '-O listings'.)",
        '7days':        "Ignore data more than 7 days old during import, and expire old records after import.",
        'units':        "Treat listing entries with 0 units as having the corresponding supply/demand price treated "
                        "as 0. This stops things like Tritium showing up where it's not available but someone was "
                        "able to sell it.",
        'bootstrap':    "Helper to 'do the right thing' and get you some data",
    }
    
    def __init__(self, tdb, tdenv):
        super().__init__(tdb, tdenv)
        
        self.dataPath = os.environ.get('TD_EDDB') or self.tdenv.tmpDir
        self.categoriesPath = Path("Category.csv")
        self.commoditiesPath = Path("Item.csv")
        self.rareItemPath = Path("RareItem.csv")
        self.shipPath = Path("Ship.csv")
        self.urlShipyard = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/shipyard.csv"
        self.FDevShipyardPath = self.tdb.dataPath / Path("FDevShipyard.csv")
        self.shipVendorPath = Path("ShipVendor.csv")
        self.stationsPath = Path("Station.csv")
        self.sysPath = Path("System.csv")
        self.upgradesPath = Path("Upgrade.csv")
        self.urlOutfitting = "https://raw.githubusercontent.com/EDCD/FDevIDs/master/outfitting.csv"
        self.FDevOutfittingPath = self.tdb.dataPath / Path("FDevOutfitting.csv")
        self.upgradeVendorPath = Path("UpgradeVendor.csv")
        self.listingsPath = Path("listings.csv")
        self.liveListingsPath = Path("listings-live.csv")
        self.pricesPath = Path("listings.prices")
    
    def now(self):
        return datetime.datetime.now().strftime('%H:%M:%S')

    def _eddblink_state_path(self) -> Path:
        """
        Single sidecar state file stored in TD_DATA (tdb.dataPath).
        This is the authoritative record of "downloaded from server" identity.
        """
        return (self.tdb.dataPath / "eddblink_state.json").resolve()

    def _load_eddblink_state(self) -> dict:
        import json

        state_path = self._eddblink_state_path()
        if not state_path.exists():
            return {"version": 1, "files": {}}

        try:
            with state_path.open("r", encoding="utf-8") as fh:
                state = json.load(fh)
            if not isinstance(state, dict):
                return {"version": 1, "files": {}}
            state.setdefault("version", 1)
            files = state.setdefault("files", {})
            if not isinstance(files, dict):
                state["files"] = {}
            return state
        except Exception:
            # Corrupt/partial JSON shouldn't brick the importer; treat as "no state"
            return {"version": 1, "files": {}}

    def _save_eddblink_state(self, state: dict) -> None:
        import json

        state_path = self._eddblink_state_path()
        state_path.parent.mkdir(parents=True, exist_ok=True)

        tmp_path = state_path.with_name(state_path.name + ".tmp")
        with tmp_path.open("w", encoding="utf-8") as fh:
            json.dump(state, fh, indent=2, sort_keys=True)
            fh.write("\n")
        tmp_path.replace(state_path)

    def _file_sha256(self, path: Path) -> str:
        import hashlib

        h = hashlib.sha256()
        with path.open("rb") as fh:
            for chunk in iter(lambda: fh.read(1024 * 1024), b""):
                h.update(chunk)
        return h.hexdigest()

    def _sanity_check_category_root(self) -> None:
        """
        Category is foundational. If it's wrong, the DB is not trustworthy.
        Minimal check: Category.category_id == 1 must be 'Metals' (case-insensitive).
        """
        rebuild_cmd = "trade import -P eddblink -O clean,skipvend"

        Session = self.tdb.Session
        try:
            with Session() as session:
                row = session.execute(
                    select(SA.Category.category_id, SA.Category.name)
                    .where(SA.Category.category_id == 1)
                ).first()
        except Exception as e:
            raise PluginException(
                "Category table check failed (missing schema or broken DB).\n"
                "This DB is not usable; rebuild your local database with:\n"
                f"    {rebuild_cmd}"
            ) from e

        if not row:
            raise PluginException(
                "Category table is missing/empty.\n"
                "This DB is not usable; rebuild your local database with:\n"
                f"    {rebuild_cmd}"
            )

        cid, name = row
        got = (str(name) if name is not None else "").strip()
        if got.lower() != "metals":
            raise PluginException(
                "Category table is corrupt: category_id=1 expected 'Metals'.\n"
                f"Got: {got!r}\n"
                "This DB is not trustworthy; rebuild your local database with:\n"
                f"    {rebuild_cmd}"
            )


    def downloadFile(self, path):
        """
        Fetch the latest dumpfile from the website based on server identity,
        not local mtime.

        Proof-of-sync is stored in TD_DATA/eddblink_state.json.
        If there's no state entry for a file, it is considered out-of-sync
        (e.g. template-copied files) and will be downloaded.
        """
        if path not in (self.liveListingsPath, self.listingsPath):
            localPath = Path(self.tdb.dataPath, path)
        else:
            localPath = Path(self.dataPath, path)

        url = BASE_URL + str(path)
        key = str(path)

        self.tdenv.NOTE("Checking for update to '{}'.", path)

        state = self._load_eddblink_state()
        files_state = state.setdefault("files", {})
        entry = files_state.get(key)

        # Local integrity check against recorded state (detect template clobber / manual edits).
        in_sync_locally = False
        if entry and localPath.exists():
            try:
                st = localPath.stat()
                if int(entry.get("size", -1)) == int(st.st_size):
                    want_sha = entry.get("sha256")
                    if want_sha:
                        got_sha = self._file_sha256(localPath)
                        if got_sha == want_sha:
                            in_sync_locally = True
                    else:
                        in_sync_locally = True
            except Exception:
                in_sync_locally = False

        # HEAD request for remote identity (ETag/Last-Modified)
        headers = {"User-Agent": "Trade-Dangerous", "Accept-Encoding": "identity"}
        try:
            response = requests.head(url, headers=headers, timeout=70)
        except Exception as e:  # pylint: disable=broad-exception-caught
            self.tdenv.WARN("Problem with download:\n    URL: {}\n    Error: {}", url, str(e))
            return False

        if not getattr(response, "ok", False):
            self.tdenv.WARN("Problem with download:\n    URL: {}\n    HTTP: {}", url, getattr(response, "status_code", "?"))
            return False

        remote_etag = response.headers.get("etag")
        remote_last_modified = response.headers.get("last-modified")
        remote_length = response.headers.get("content-length")

        dump_mod_time = None
        if remote_last_modified:
            try:
                dump_mod_time = parsedate_to_datetime(remote_last_modified).timestamp()
            except Exception:
                dump_mod_time = None

        # If we have a prior server-proven state AND local file matches that state,
        # we can skip downloading when remote identity matches.
        if entry and in_sync_locally:
            # Prefer ETag when available; else fall back to Last-Modified.
            if remote_etag and entry.get("etag") == remote_etag:
                self.tdenv.DEBUG0("'{}': Remote ETag matches state; no download.", path)
                return False
            if (not remote_etag) and remote_last_modified and entry.get("last_modified") == remote_last_modified:
                self.tdenv.DEBUG0("'{}': Remote Last-Modified matches state; no download.", path)
                return False

        # If state is missing, or local doesn't match recorded state, or remote identity differs -> download.
        self.tdenv.NOTE("Downloading file '{}'.", path)
        transfers.download(self.tdenv, url, localPath, chunkSize=16384, length=remote_length)

        # Change timestamps on the file to match the server (human convenience only)
        if dump_mod_time is not None:
            try:
                os.utime(localPath, (dump_mod_time, dump_mod_time))
            except Exception:
                pass

        # Update sync state (stored in TD_DATA regardless of localPath location)
        try:
            st = localPath.stat()
            new_entry = {
                "url": url,
                "local_path": str(localPath.resolve()),
                "etag": remote_etag,
                "last_modified": remote_last_modified,
                "content_length": remote_length,
                "downloaded_at": datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat(),
                "size": int(st.st_size),
            }

            # Hash only the small “truth-critical” files (cheap + detects template clobber cleanly).
            if key in ("Category.csv", "RareItem.csv", "Item.csv"):
                new_entry["sha256"] = self._file_sha256(localPath)

            files_state[key] = new_entry
            self._save_eddblink_state(state)
        except Exception:
            # State failures must not make downloads fail.
            pass

        return True
    
    def purgeSystems(self):
        """
        Purges systems from the System table that do not have any stations claiming to be in them.
        Keeps table from becoming too large because of fleet carriers moving to unpopulated systems.
        """
        self.tdenv.NOTE("Purging Systems with no stations: Start time = {}", self.now())
        
        Session = self.tdb.Session
        with Session.begin() as session:
            subq = select(SA.Station.system_id).where(SA.Station.system_id == SA.System.system_id)
            stmt = delete(SA.System).where(~exists(subq))
            session.execute(stmt)
        
        self.tdenv.NOTE("Finished purging Systems. End time = {}", self.now())
    
    def importListings(self, listings_file):
        """
        Updates the market data (StationItem) using `listings_file`.
        
        Rules:
          - If a row doesn't exist in DB → insert (copy CSV exactly).
          - If it exists → update only when CSV.modified > DB.modified.
          - If CSV.modified <= DB.modified → do nothing (no field changes).
        """
        listings_path = Path(self.dataPath, listings_file).absolute()
        from_live = listings_path != Path(self.dataPath, self.listingsPath).absolute()
        
        self.tdenv.NOTE("Checking listings")
        total = _count_listing_entries(self.tdenv, listings_path)
        if not total:
            self.tdenv.NOTE("No listings")
            return
        
        self.tdenv.NOTE(
            "Processing market data from {}: Start time = {}, Live = {}",
            listings_file, self.now(), from_live
        )
        
        Session = self.tdb.Session
        
        # Prefetch item/station IDs for early filtering
        with Session.begin() as session:
            item_lookup = _make_item_id_lookup(self.tdenv, session)
            station_lookup = _make_station_id_lookup(self.tdenv, session)
        
        self.tdenv.DEBUG0("Processing entries...")
        
        with pbar.Progress(total, 40, label="Processing", style=pbar.LongRunningCountBar) as prog, \
               listings_path.open("r", encoding="utf-8", errors="ignore") as fh, \
               Session() as session:
            
            token = begin_bulk_mode(session, profile="eddblink", phase="incremental")
            try:
                commit_batch = get_import_batch_size(session, profile="eddblink")
                execute_batch = commit_batch or 10000  # cap statement size even if single final commit
                
                # Upsert: keys + guarded fields (including from_live), guarded by 'modified'
                table = SA.StationItem.__table__
                key_cols = ("station_id", "item_id")
                update_cols = (
                    "demand_price", "demand_units", "demand_level",
                    "supply_price", "supply_units", "supply_level",
                    "from_live",
                )
                upsert = get_upsert_fn(
                    session,
                    table,
                    key_cols=key_cols,
                    update_cols=update_cols,
                    modified_col="modified",
                    always_update=(),   # IMPORTANT: no unconditional updates
                )
                
                batch_rows = []
                since_commit = 0
                
                # optimize away millions of lookups
                increment = prog.increment
                
                def bump_progress():
                    increment(1)
                
                from_timestamp = datetime.datetime.fromtimestamp
                utc = datetime.timezone.utc
                from_live_val = int(from_live)
                week_in_seconds = 7 * 24 * 60 * 60
                time_cutoff = 0 if not self.getOption("7days") else time.time() - week_in_seconds
                squelch_zero_units = self.getOption("units")
                
                # Columns:
                #
                #   id, station_id, commodity_id, supply, supply_bracket, buy_price, sell_price, demand, demand_bracket, collected_at
                #   0   1           2             3       4               5          6           7       8               9
                reader = iter(csv.reader(fh))
                headers = next(reader)
                expect_headers = [
                    "id", "station_id", "commodity_id",
                    "supply", "supply_bracket", "buy_price",
                    "sell_price", "demand", "demand_bracket",
                    "collected_at"
                ]
                if headers[:10] != expect_headers:
                    raise TradeException(
                        f"incompatible csv field organization in {listings_path}. "
                        f"expected {expect_headers}; got {headers}"
                    )
                
                for listing in reader:
                    bump_progress()
                    try:
                        if squelch_zero_units:
                            if listing[3] == "0":
                                listing[3] = listing[4] = listing[5] = "0"
                            if listing[7] == "0":
                                listing[6] = listing[7] = listing[8] = "0"
                        
                        # Do the cheapest skip-check first
                        if listing[5] == "0" and listing[6] == "0":
                            continue
                        
                        # Cheap numeric condition
                        listing_time = int(listing[9])
                        if listing_time < time_cutoff:
                            continue
                        
                        station_id = int(listing[1])
                        if station_id not in station_lookup:
                            continue
                        
                        item_id = int(listing[2])
                        if item_id not in item_lookup:
                            continue  # skip rare items (not in Item table)
                        
                        dt_listing_time = from_timestamp(listing_time, utc)
                        
                        row = {
                            "station_id":   station_id,
                            "item_id":      item_id,
                            "modified":     dt_listing_time,   # guard column
                            "from_live":    from_live_val,     # copied exactly when updating/inserting
                            "supply_units": int(listing[3]),
                            "supply_level": int(listing[4]),
                            "supply_price": int(listing[5]),
                            "demand_price": int(listing[6]),
                            "demand_units": int(listing[7]),
                            "demand_level": int(listing[8]),
                        }
                        batch_rows += [row]
                        since_commit += 1
                        
                        if len(batch_rows) >= execute_batch:
                            upsert(batch_rows)
                            batch_rows[:] = []  # in-place clear without lookup
                        
                        if commit_batch and since_commit >= commit_batch:
                            session.commit()
                            since_commit = 0
                    
                    except Exception as e:  # pylint: disable=broad-exception-caught
                        self.tdenv.WARN("Bad listing row (skipped): {}  error: {}", listing, e)
                        continue
                
                if batch_rows:
                    upsert(batch_rows)
                    batch_rows[:] = []  # in-place clear
                
                session.commit()
            
            finally:
                end_bulk_mode(session, token)
        
        # with pbar.Progress(1, 40, prefix="Saving"):
        #     pass
        
        if self.getOption("7days"):
            # This is a gimmick for first-time pruning: instead of trying to delete
            # years of old data, do it a piece at a time. It gives the progress bar
            # some movement.
            expirations = [360, 330, 300, 270, 240, 210, 180, 150, 120, 90, 60, 30, 21, 14, 7]
            with pbar.Progress(len(expirations) + 1, 40, 1, label="Expiring", style=pbar.LongRunningCountBar) as prog, \
                    Session.begin() as session:
                for expiration in expirations:
                    session.execute(text(f"DELETE FROM StationItem WHERE modified < datetime('now', '-{expiration} days')"))
                    prog.increment(1)
        
        if self.getOption("optimize"):
            with pbar.Progress(0, 40, label="Optimizing", style=pbar.ElapsedBar) as prog:
                if self.tdb.engine.dialect.name == "sqlite":
                    with Session.begin() as session:
                        session.execute(text("VACUUM"))
        
        self.tdenv.NOTE("Finished processing market data. End time = {}", self.now())
    
    def _refresh_dump_tables(self, table_jobs: list[tuple[str, Path]]) -> None:
        """Upsert-refresh (table_name, csv_path) jobs into the live ORM database,
        with a proper row-count progress bar.

        Note: RareItem is rebuilt (wiped then re-imported) whenever it is refreshed.
        This avoids UNIQUE(name) collisions caused by historical PK drift / template-era imports.
        """
        if not table_jobs:
            return

        # Local import to avoid plugin import-order headaches.
        from tradedangerous import cache as td_cache

        Session = self.tdb.Session
        with Session() as session:
            with pbar.Progress(
                max_value=len(table_jobs) + 1,
                prefix="Upserting",
                width=25,
                style=pbar.CountingBar,
            ) as prog:
                for table_name, import_path in table_jobs:
                    import_lines = file_line_count(import_path, missing_ok=True)
                    with prog.sub_task(
                        max_value=import_lines,
                        description=table_name,
                    ) as child:
                        prog.increment(value=1)
                        call_args = {"task": child, "advance": 1}
                        try:
                            # RareItem: rebuild contents on refresh to avoid uq_rareitem_name collisions
                            # when existing DB has same names under different rare_id values.
                            if table_name == "RareItem":
                                session.execute(delete(SA.RareItem))

                            td_cache.processImportFile(
                                self.tdenv,
                                session,
                                import_path,
                                table_name,
                                line_callback=prog.update_task,
                                call_args=call_args,
                            )
                            session.commit()
                        except FileNotFoundError:
                            self.tdenv.WARN("Missing import file for {}: {}", table_name, import_path)
                        except StopIteration:
                            self.tdenv.NOTE(
                                "{} exists but is empty. Remove it or add the column definition line.",
                                import_path,
                            )

                prog.increment(1)


    def run(self):
        """
        EDDN/EDDB link importer.

        Refactored DB flow:
          - No dialect-specific logic in the plugin.
          - Preflight uses TradeDB.reloadCache() (which centralizes sanity via lifecycle.ensure_fresh_db).
          - For '--clean' → do a single full rebuild with the RareItem dance.
          - Otherwise, if static CSVs changed → upsert-refresh only those tables (no drop/recreate).
          - Listings import unchanged.
        """
        self.tdenv.ignoreUnknown = True
        self.tdb.dataPath.mkdir(parents=True, exist_ok=True)

        # Enable 'listings' by default unless other explicit options are present
        default = True
        for option in self.options:
            if option not in ('force', 'skipvend', 'purge', '7days', 'units'):
                default = False
        if default:
            self.options["listings"] = True

        if self.getOption("bootstrap"):
            self.tdenv.NOTE("[bold][blue]bootstrap: Greetings, Commander!")
            self.tdenv.NOTE(
                "[yellow]This first-time import might take several minutes or longer, "
                "it ensures your database is up to date with current EDDBLink System, Station, and Item tables "
                "as well as trade listings for the last 7 days.")
            self.tdenv.NOTE(
                "[yellow]You can run this same command later to import updates - which should be much faster, "
                "or `trade import -P eddblink -O 7days,skipvend`.")
            self.tdenv.NOTE(
                "[yellow]To contribute your own discoveries to market data, consider running the "
                "Elite Dangerous Market Connector while playing.")
            for child in ["system", "station", "item", "listings", "skipvend", "7days"]:
                self.options[child] = True

        # Check if database already exists and enable `clean` if not.
        if lifecycle.is_empty(self.tdb.engine):
            self.options["clean"] = True

        if self.getOption("clean"):
            # Remove CSVs so downloads become the new source of truth
            for name in [
                "Category", "Item", "RareItem",
                "Ship", "ShipVendor",
                "Station", "System",
                "Upgrade", "UpgradeVendor",
                "FDevShipyard", "FDevOutfitting",
            ]:
                f = self.tdb.dataPath / f"{name}.csv"
                try:
                    os.remove(str(f))
                except FileNotFoundError:
                    pass

            # Remove eddblink sync-state (sidecar) so templates never "win"
            try:
                os.remove(str(self._eddblink_state_path()))
            except FileNotFoundError:
                pass

            # Remove .prices (DEPRECATED)
            try:
                os.remove(str(self.tdb.dataPath / "TradeDangerous.prices"))
            except FileNotFoundError:
                pass

            self.options["all"] = True
            self.options["force"] = True
        else:
            # Category is foundational; if it's wrong, this DB is not trustworthy.
            # Hard-fail and force rebuild rather than attempting to "refresh" it.
            self._sanity_check_category_root()

        # Select which options will be updated
        if self.getOption("listings"):
            self.options["item"] = True
            self.options["station"] = True

        if self.getOption("shipvend"):
            self.options["ship"] = True
            self.options["station"] = True

        if self.getOption("upvend"):
            self.options["upgrade"] = True
            self.options["station"] = True

        if self.getOption("item"):
            self.options["station"] = True

        if self.getOption("rare"):
            self.options["station"] = True

        if self.getOption("station"):
            self.options["system"] = True

        if self.getOption("all"):
            self.options["item"] = True
            self.options["rare"] = True
            self.options["ship"] = True
            self.options["shipvend"] = True
            self.options["station"] = True
            self.options["system"] = True
            self.options["upgrade"] = True
            self.options["upvend"] = True
            self.options["listings"] = True

        if self.getOption("solo"):
            self.options["listings"] = False
            self.options["skipvend"] = True

        if self.getOption("skipvend"):
            self.options["shipvend"] = False
            self.options["upvend"] = False

        # Download required files and decide which tables need upsert-refresh.
        force = self.getOption("force")

        upgrade_changed = False
        ship_changed = False
        rare_changed = False
        shipvend_changed = False
        upvend_changed = False
        system_changed = False
        station_changed = False
        category_changed = False
        item_changed = False

        # FDev bridge CSVs are treated as "changed" when we re-download them.
        fdev_shipyard_changed = False
        fdev_outfitting_changed = False

        if self.getOption("upgrade"):
            upgrade_changed = self.downloadFile(self.upgradesPath) or force
            if upgrade_changed:
                transfers.download(self.tdenv, self.urlOutfitting, self.FDevOutfittingPath)
                fdev_outfitting_changed = True

        if self.getOption("ship"):
            ship_changed = self.downloadFile(self.shipPath) or force
            if ship_changed:
                transfers.download(self.tdenv, self.urlShipyard, self.FDevShipyardPath)
                fdev_shipyard_changed = True

        if self.getOption("rare"):
            rare_changed = self.downloadFile(self.rareItemPath) or force

        if self.getOption("shipvend"):
            shipvend_changed = self.downloadFile(self.shipVendorPath) or force

        if self.getOption("upvend"):
            upvend_changed = self.downloadFile(self.upgradeVendorPath) or force

        if self.getOption("system"):
            system_changed = self.downloadFile(self.sysPath) or force

        if self.getOption("station"):
            station_changed = self.downloadFile(self.stationsPath) or force

        if self.getOption("item"):
            item_changed = self.downloadFile(self.commoditiesPath) or force
            # Category can change independently; always check when item option is active.
            category_changed = self.downloadFile(self.categoriesPath) or force

        # If any of the non-listings tables changed, ensure DB is fresh and then upsert-refresh.
        build_cache = any([
            upgrade_changed, ship_changed, rare_changed,
            shipvend_changed, upvend_changed,
            system_changed, station_changed,
            category_changed, item_changed,
            fdev_shipyard_changed, fdev_outfitting_changed,
        ])

        if build_cache:
            if self.getOption("clean"):
                # "clean" must mean clean for all backends:
                #   - sqlite  → rotate/recreate DB file
                #   - mariadb → drop+recreate tables (NOT the database)
                self.tdenv.NOTE("NOTE: --clean requested; resetting database schema.")
                self.tdb.close()
                lifecycle.reset_db(
                    self.tdb.engine,
                    db_path=self.tdb.dbPath,
                    sql_path=self.tdb.sqlPath,
                )
            else:
                # Ensure schema exists and is sane (may rebuild on first run).
                self.tdb.close()
                self.tdb.reloadCache()

            if self.tdb.engine.dialect.name == "sqlite":
                # kfsone: see https://sqlite.org/pragma.html#pragma_optimize
                self.tdb.Session().execute(text("PRAGMA optimize=0x10002"))

            # Upsert-refresh tables in dependency order.
            jobs: list[tuple[str, Path]] = []

            if system_changed:
                jobs.append(("System", (self.tdb.dataPath / self.sysPath).resolve()))

            if station_changed:
                jobs.append(("Station", (self.tdb.dataPath / self.stationsPath).resolve()))

            if category_changed or item_changed:
                jobs.append(("Category", (self.tdb.dataPath / self.categoriesPath).resolve()))
                jobs.append(("Item", (self.tdb.dataPath / self.commoditiesPath).resolve()))

            if ship_changed:
                jobs.append(("Ship", (self.tdb.dataPath / self.shipPath).resolve()))
            if fdev_shipyard_changed:
                jobs.append(("FDevShipyard", self.FDevShipyardPath.resolve()))

            if upgrade_changed:
                jobs.append(("Upgrade", (self.tdb.dataPath / self.upgradesPath).resolve()))
            if fdev_outfitting_changed:
                jobs.append(("FDevOutfitting", self.FDevOutfittingPath.resolve()))

            if shipvend_changed:
                jobs.append(("ShipVendor", (self.tdb.dataPath / self.shipVendorPath).resolve()))

            if upvend_changed:
                jobs.append(("UpgradeVendor", (self.tdb.dataPath / self.upgradeVendorPath).resolve()))

            if rare_changed:
                jobs.append(("RareItem", (self.tdb.dataPath / self.rareItemPath).resolve()))

            self._refresh_dump_tables(jobs)
            self.tdb.close()

        if self.getOption("purge"):
            self.purgeSystems()

        # Listings import (prices)
        if self.getOption("listings"):
            if self.downloadFile(self.listingsPath) or force:
                self.importListings(self.listingsPath)
            if self.downloadFile(self.liveListingsPath) or force:
                self.importListings(self.liveListingsPath)

        if self.tdb.engine.dialect.name == "sqlite":
            with self.tdb.Session.begin() as session:
                if self.getOption("optimize"):
                    with bench("Vacuum and optimize", self.tdenv):
                        session.execute(text("VACUUM"))
                        # This is a very aggressive analyze/optimize pass
                        session.execute(text("ANALYZE"))
                else:
                    with bench("DB Tuning", self.tdenv):
                        session.execute(text("PRAGMA optimize"))
                    self.tdenv.INFO("Use --opt=optimize periodically for better query performance")
                    
        self.tdenv.NOTE("Import completed.")
        
        return False


    def finish(self):
        """ override the base class 'finish' method """
        # We expect to return 'False' from run, so if this is called, something went horribly wrong;
        # if this gets reached, someone added a bad return to run().
        self.tdenv.WARN("Internal error: plugin's finish() method was reached")
        return False



# --------------------------------------------------------------------
# Copyright (C) Oliver 'kfsone' Smith 2014 <oliver@kfs.org>:
# Copyright (C) Bernd 'Gazelle' Gollesch 2016, 2017
# Copyright (C) Stefan 'Tromador' Morrell 2025
# Copyright (C) Jonathan 'eyeonus' Jones 2018-2025
#
# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: Modules :: Generate TradeDangerous.prices

import sys
from sqlalchemy.orm import Session
from .db import orm_models as SA
from .tradeexcept import TradeException


class Element:      # TODO: enum?
    basic     = 1 << 0
    supply    = 1 << 1
    timestamp = 1 << 2
    full      = basic | supply | timestamp
    blanks    = 1 << 31


######################################################################
# Main

def dumpPrices(
    session: Session,      # SQLAlchemy session
    elementMask,           # which columns to output
    stationID=None,        # limits to one station
    file=None,             # file handle to write to
    defaultZero=False,
    debug=0,
):
    """
    Generate a prices list using data from the DB.
    If stationID is not None, only the specified station is dumped.
    If file is not None, outputs to the given file handle.
    """
    
    withTimes = elementMask & Element.timestamp
    
    # ORM queries to build lookup dicts
    systems = dict(
        session.query(SA.System.system_id, SA.System.name).all()
    )
    
    stations = {
        ID: [name, systems.get(sysID)]
        for ID, name, sysID in session.query(
            SA.Station.station_id, SA.Station.name, SA.Station.system_id
        ).all()
    }
    
    categories = dict(
        session.query(SA.Category.category_id, SA.Category.name).all()
    )
    
    items = {
        ID: [name, catID, categories[catID]]
        for ID, name, catID in session.query(
            SA.Item.item_id, SA.Item.name, SA.Item.category_id
        ).all()
    }
    
    # find longest item name (for formatting)
    longestName = max(items.values(), key=lambda ent: len(ent[0]))
    longestNameLen = len(longestName[0])
    
    defaultDemandVal = 0 if defaultZero else -1
    
    # Build the main query
    q = (
        session.query(
            SA.StationItem.station_id,
            SA.Item.item_id,
            SA.StationItem.demand_price,
            SA.StationItem.supply_price,
            SA.StationItem.demand_units,
            SA.StationItem.demand_level,
            SA.StationItem.supply_units,
            SA.StationItem.supply_level,
            SA.StationItem.modified,
            SA.Item.name,
            SA.Item.category_id,
            SA.Category.name.label("category_name"),
            SA.Station.name.label("station_name"),
            SA.System.name.label("system_name"),
        )
        .join(SA.Item, SA.Item.item_id == SA.StationItem.item_id)
        .join(SA.Category, SA.Category.category_id == SA.Item.category_id)
        .join(SA.Station, SA.Station.station_id == SA.StationItem.station_id)
        .join(SA.System, SA.System.system_id == SA.Station.system_id)
        .order_by(SA.Station.station_id, SA.Category.name, SA.Item.ui_order)
    )
    
    if stationID:
        q = q.filter(SA.StationItem.station_id == stationID)
    
    # Set up output
    if not file:
        file = sys.stdout
    
    if stationID:
        stationSet = str(stations[stationID])
    else:
        stationSet = "ALL Systems/Stations"
    
    file.write(
        "# TradeDangerous prices for {}\n"
        "\n"
        "# REMOVE ITEMS THAT DON'T APPEAR IN THE UI\n"
        "# ORDER IS REMEMBERED: Move items around within categories "
        "to match the game UI\n"
        "\n"
        "# File syntax:\n"
        "# <item name> <sell> <buy> [<demand> <supply> [<timestamp>]]\n"
        "#   Use '?' for demand/supply when you don't know/care,\n"
        "#   Use '-' for demand/supply to indicate unavailable,\n"
        "#   Otherwise use a number followed by L, M or H, e.g.\n"
        "#     1L, 23M or 30000H\n"
        "# If you omit the timestamp, the current time will be used when "
        "the file is loaded.\n"
        "\n".format(stationSet)
    )
    
    levelDesc = "?0LMH"
    maxCrWidth = 7
    levelWidth = 9
    
    outFmt = (
        "      {{:<{width}}}"
        " {{:>{crwidth}}}"
        " {{:>{crwidth}}}"
        "  {{:>{lvlwidth}}}"
        " {{:>{lvlwidth}}}".format(
            width=longestNameLen,
            crwidth=maxCrWidth,
            lvlwidth=levelWidth,
        )
    )
    if withTimes:
        outFmt += "  {}"
    outFmt += "\n"
    header = outFmt.format(
        "Item Name",
        "SellCr", "BuyCr",
        "Demand", "Supply",
        "Timestamp",
    )
    file.write('#' + header[1:])
    
    naIQL = "-"
    unkIQL = "?"
    defIQL = "?" if not defaultZero else "-"
    
    # Main loop — stream results instead of preloading
    output = ""
    lastStn, lastCat = None, None
    for row in q.yield_per(1000).execution_options(stream_results=True):
        stnID = row.station_id
        station = row.station_name
        system = row.system_name
        item = row.name
        catID = row.category_id
        category = row.category_name
        
        # Guard against bad system names
        if not system:
            raise TradeException(
                f"Station {station} (ID {stnID}) is linked to a system with no name."
            )
        
        if stnID != lastStn:
            file.write(output)
            output = f"\n\n@ {system.upper()}/{station}\n"
            lastStn = stnID
            lastCat = None
        
        if catID != lastCat:
            output += f"   + {category}\n"
            lastCat = catID
        
        demandCr = row.demand_price or 0
        supplyCr = row.supply_price or 0
        demandUnits = row.demand_units or defaultDemandVal
        demandLevel = row.demand_level or defaultDemandVal
        supplyUnits = row.supply_units or defaultDemandVal
        supplyLevel = row.supply_level or defaultDemandVal
        
        # Demand/supply formatting
        if supplyCr > 0:
            demandStr = defIQL if demandCr <= 0 else unkIQL
            supplyStr = (
                naIQL if supplyLevel == 0
                else (f"{supplyUnits if supplyUnits >= 0 else '?'}{levelDesc[supplyLevel+1]}")
            )
        else:
            demandStr = (
                naIQL if demandCr == 0 or demandLevel == 0
                else (f"{demandUnits if demandUnits >= 0 else '?'}{levelDesc[demandLevel+1]}")
            )
            supplyStr = naIQL
        
        modified = row.modified or ""
        output += outFmt.format(item, demandCr, supplyCr, demandStr, supplyStr, modified)
    
    file.write(output)



from ..csvexport import exportTableToFile
from .parsing import ParseArgument, MutuallyExclusiveGroup
from .exceptions import CommandLineError
from pathlib import Path

######################################################################
# TradeDangerous :: Commands :: Export
#
# Generate the CSV files for the master data of the database.
#
######################################################################
# CAUTION: If the database structure gets changed this script might
#          need some corrections.
######################################################################

######################################################################
# Parser config

help='CSV exporter for TradeDangerous database.'
name='export'
epilog=(
        "CAUTION: If you don't specify a different path, the current "
        "CSV files in the data directory will be overwritten with "
        "the current content of the database.\n "
        "If you have changed any CSV file and didn't rebuild the "
        "database, they will be lost.\n "
        "Use the 'buildcache' command first to rebuild the database."
)
wantsTradeDB=False           # because we don't want the DB to be rebuild
arguments = [
]
switches = [
    ParseArgument('--path',
            help="Specify a different save location of the CSV files than the default.",
            type=str,
            default=None
        ),
    MutuallyExclusiveGroup(
        ParseArgument('--tables', "-T",
                help='Specify comma separated tablenames to export.',
                metavar='TABLE[,TABLE,...]',
                type=str,
                default=None
            ),
        ParseArgument('--all-tables',
                help='Include the price tables for export.',
                dest='allTables',
                action='store_true',
                default=False
            ),
        ),
    ParseArgument('--delete-empty',
            help='Delete CSV files without content.',
            dest='deleteEmpty',
            action='store_true',
            default=False
        ),
]

######################################################################
# Perform query and populate result set

def run(results, cmdenv, tdb):
    """
    Backend-neutral export of DB tables to CSV.
    
    Changes:
      * Use tradedangerous.db.lifecycle.ensure_fresh_db(rebuild=False) to verify a usable DB
        without rebuilding (works for SQLite and MariaDB).
      * Backend-aware announcement of the source DB (file path for SQLite, DSN for others).
      * Table enumeration via SQLAlchemy inspector (no sqlite_master, no COLLATE quirks).
    """
    # --- Sanity check the database without rebuilding (works for both backends) ---
    from tradedangerous.db.lifecycle import ensure_fresh_db  # local import avoids import-time tangles
    summary = ensure_fresh_db(
        backend=getattr(tdb.engine, "dialect", None).name if getattr(tdb, "engine", None) else "unknown",
        engine=getattr(tdb, "engine", None),
        data_dir=tdb.dataPath,
        metadata=None,
        mode="auto",
        rebuild=False,   # IMPORTANT: never rebuild from here; just report health
    )
    if summary.get("sane") != "Y":
        reason = summary.get("reason", "unknown")
        raise CommandLineError(
            f"Database is not initialized/healthy (reason: {reason}). "
            "Use 'buildcache' or an importer to (re)build it."
        )
    
    # --- Determine export target directory (same behavior as before) ---
    exportPath = Path(cmdenv.path) if cmdenv.path else Path(tdb.dataDir)
    if not exportPath.is_dir():
        raise CommandLineError("Save location '{}' not found.".format(str(exportPath)))
    
    # --- Announce which DB we will read from, backend-aware ---
    try:
        dialect = tdb.engine.dialect.name
        if dialect == "sqlite":
            source_label = f"SQLite file '{tdb.dbPath}'"
        else:
            # Hide password in DSN
            source_label = f"{dialect} @ {tdb.engine.url.render_as_string(hide_password=True)}"
    except Exception:
        source_label = str(getattr(tdb, "dbPath", "Unknown DB"))
    cmdenv.NOTE("Using database {}", source_label)
    
    # --- Enumerate tables using SQLAlchemy inspector (backend-neutral) ---
    from sqlalchemy import inspect
    inspector = inspect(tdb.engine)
    all_tables = inspector.get_table_names()  # current schema / database
    
    # Optional ignore list (preserve legacy default: skip StationItem unless --all-tables)
    ignoreList = []
    if not getattr(cmdenv, "allTables", False):
        ignoreList.append("StationItem")
    
    # --tables filtering (case-insensitive, like old COLLATE NOCASE)
    if getattr(cmdenv, "tables", None):
        requested = [t.strip() for t in cmdenv.tables.split(",") if t.strip()]
        lower_map = {t.lower(): t for t in all_tables}
        resolved = []
        for name in requested:
            found = lower_map.get(name.lower())
            if found:
                resolved.append(found)
            else:
                cmdenv.NOTE("Requested table '{}' not found; skipping", name)
        table_list = sorted(set(resolved))
    else:
        table_list = sorted(set(all_tables))
    
    # --- Export each table via csvexport (already refactored elsewhere) ---
    for tableName in table_list:
        if tableName in ignoreList:
            cmdenv.NOTE("Ignore Table '{table}'", table=tableName)
            continue
        
        cmdenv.NOTE("Export Table '{table}'", table=tableName)
        
        lineCount, filePath = exportTableToFile(tdb, cmdenv, tableName, exportPath)
        
        # Optionally delete empty CSVs
        if getattr(cmdenv, "deleteEmpty", False) and lineCount == 0:
            try:
                filePath.unlink(missing_ok=True)
                cmdenv.DEBUG0("Delete empty file {file}", file=filePath)
            except Exception as e:
                cmdenv.DEBUG0("Failed to delete empty file {file}: {err}", file=filePath, err=e)
    
    return False  # we've handled everything



#!/usr/bin/env python3
# --------------------------------------------------------------------
# Copyright (C) Jonathan 'eyeonus' Jones 2018-2022

# You are free to use, redistribute, or even print and eat a copy of
# this software so long as you include this copyright notice.
# I guarantee there is at least one bug neither of us knew about.
# --------------------------------------------------------------------
# TradeDangerous :: GUI App :: Main Module

# Where all the graphics happens. Uses TD CLI under the hood.

# Current features:
# ----------------
# Drop-down list of all available TD commands
# Fully populated list of all arguments and switches for each command
# Automatic setting of default value for the above which have one
# Procedural generation of all above for future proofing in the
#   event of new import plugins, switches, arguments, commands(?)
# RAM resident save-state: altered values retain new value while main
#    window remains open

# Planned features:
# ----------------
# User-defined initial values AKA tdrc files (.tdrc_run, .tdrc_trade, ..)
# Profiles AKA fromfiles (+ship1, +ship2, ..)
# Select-able output text
# graphical render of results
# send results to new window
# individual always-on-top for every window
# Data retrieval from CMDR's journal

import os
import sys
import threading
import tkinter as tk
import traceback
from pathlib import Path
from tkinter import ttk, filedialog

from . import commands
from . import plugins
from . import tradedb
from .commands import exceptions
from .plugins import PluginException
from .version import __version__


import tkinter as tk
from tkinter import ttk, scrolledtext

class TDWidget:
    """
    Tkinter widget wrapper supporting appJar-style context management.
    """
    
    def __init__(self, name=None, widget_type='frame', parent=None, row=0, column=0,
                 rowspan=1, columnspan=1, sticky='nw', text='', values=None,
                 command=None, width=None, height=None, **kwargs):
        self.kwargs = kwargs
        self.name = name
        self.widget_type = widget_type
        self.children = []
        
        # Parent widget
        container = parent.widget if isinstance(parent, TDWidget) else parent
        
        # Create underlying Tk widget
        if widget_type == 'frame':
            self.widget = tk.Frame(container, width=width, height=height)
        elif widget_type == 'label':
            self.widget = tk.Label(container, text=text, width=width)
        elif widget_type == 'button':
            self.widget = tk.Button(container, text=text, command=command)
        elif widget_type == 'entry':
            self.var = tk.StringVar(value=text)
            self.widget = tk.Entry(container, textvariable=self.var, width=width)
        elif widget_type == 'combo':
            self.var = tk.StringVar(value=text)
            self.widget = ttk.Combobox(container, values=values or [], textvariable=self.var, width=width)
        elif widget_type == 'spin':
            self.var = tk.IntVar(value=text)
            self.widget = tk.Spinbox(container, from_=0, to=values or 100, textvariable=self.var, width=width)
        elif widget_type == 'scrolledtext':
            self.widget = scrolledtext.ScrolledText(container, width=width, height=height)
        elif widget_type == 'notebook':
            self.widget = ttk.Notebook(container)
            self.tabs = {}
        else:
            raise ValueError(f"Unknown widget_type: {widget_type}")
        
        # Parent-child management
        self.parent = parent
        if parent and isinstance(parent, TDWidget):
            parent.children.append(self)
        
        # Grid placement
        self.widget.grid(row=row, column=column, rowspan=rowspan, columnspan=columnspan, sticky=sticky)
        self.row, self.column = row, column
    
    # --- Proxy methods ---
    def config(self, **kwargs):
        self.widget.config(**kwargs)
    
    def set(self, value, callFunction=True):
        if hasattr(self, 'var'):
            self.var.set(value)
            if callFunction and self.widget_type == 'combo':
                if hasattr(self.widget, 'event_generate'):
                    self.widget.event_generate("<<ComboboxSelected>>")
    
    def get(self):
        if hasattr(self, 'var'):
            return self.var.get()
        return None
    
    def add_child(self, child):
        self.children.append(child)
    
    # --- Context manager support ---
    def __enter__(self):
        # Return self so "with TDWidget(...) as w:" works
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        # Nothing special to clean up
        pass
    
    # --- Notebook/tab support ---
    def tab(self, tab_name):
        """Create a new tab inside a notebook TDWidget"""
        if self.widget_type != 'notebook':
            raise RuntimeError("tab() can only be called on notebook widgets")
        frame = tk.Frame(self.widget)
        self.widget.add(frame, text=tab_name)
        self.tabs[tab_name] = frame
        return TDWidget(parent=self, widget_type='frame', row=0, column=0, columnspan=1, widget=frame)


# Plugins available to the 'import' command are stored here.
# The list is populated by scanning the plugin folder directly,
# so it updates automagically at start as plugins are added or removed.

# Any other command with available plugins must have a similar list.
importPlugs = [ plug.name[0:plug.name.find('_plug.py')]
             for plug in os.scandir(sys.modules['tradedangerous.plugins'].__path__[0])
             if plug.name.endswith("_plug.py")
             ]

widgets = {}

# All available commands
Commands = ['help'] + [ cmd for cmd, module in sorted(commands.commandIndex.items()) ]
# Used to run TD cli commands.
cmdenv = commands.CommandIndex().parse
# 'help' output, required & optional arguments, for each command
cmdHelp = {}

# Path of the database
dbS = str(Path((os.environ.get('TD_DATA') or os.path.join(os.getcwd(), 'data')) / Path('TradeDangerous.db')))
# Path of the current working directory
cwdS = str(Path(os.getcwd()))

def changeCWD():
    """
    Opens a folder select dialog for choosing the current working directory.
    """
    cwd = filedialog.askdirectory(title = "Select the top-level folder for TD to work in...",
                                  initialdir = argVals['--cwd'])
    # cwd = win.directoryBox("Select the top-level folder for TD to work in...", dirName = argVals['--cwd'])
    if cwd:
        argVals['--cwd'] = str(Path(cwd))
    widgets['cwd']['text'] = argVals['--cwd']

def changeDB():
    """
    Opens a file select dialog for choosing the database file.
    """
    db = filedialog.askopenfilename(title = "Select the TD database file to use...",
                                    initialdir = str(Path(argVals['--db']).parent),
                                    filetypes = [('Data Base File', '*.db')])
    if db:
        argVals['--db'] = str(Path(db))
    widgets['db']['text'] = argVals['--db']        


# A dict of all arguments in TD (mostly auto-generated)
# Manually add the global arguments for now, maybe figure out how to auto-populate them as well.
allArgs = {
    '--debug': { 'help': 'Enable/raise level of diagnostic output.',
                'default':  0, 'required': False, 'action': 'count',
                'widget': {'type':'combo', 'values': ['', '-w', '-ww', '-www']}
                },
    '--detail':{ 'help': 'Increase level of detail in output.',
                'default': 0, 'required': False, 'action': 'count',
                'excludes': ['--quiet'], 'widget': {'type':'combo', 'values': ['', '-v', '-vv', '-vvv']}
                },
    '--quiet':{ 'help': 'Reduce level of detail in output.',
               'default': 0, 'required': False, 'action': 'count',
               'excludes': ['--detail'], 'widget': {'type':'combo', 'values': ['', '-q', '-qq', '-qqq']}
               },
    '--db':{ 'help': 'Specify location of the SQLite database.',
            'default': dbS, 'dest': 'dbFilename', 'type': str,
            'widget': {'type':'button', 'func':changeDB}
            },
    '--cwd':{ 'help': 'Change the working directory file accesses are made from.',
             'default': cwdS, 'type': str, 'required': False,
             'widget': {'type':'button', 'func':changeCWD}
             },
    '--link-ly':{ 'help': 'Maximum lightyears between systems to be considered linked.',
                 'default': '30',
                 'widget': {'type':'entry', 'sub':'numeric'}
                 }
    }

# Used to save the value of the arguments.
argVals = {'--debug': '',
           '--detail': '',
           '--quiet': '',
           '--db': dbS,
           '--cwd': cwdS,
           '--link-ly': '30'
           }

def buildArgDicts():
    """
    Procedurally generates the contents of allArgs and argVals
    """
    try:
        cmdenv(['help'])
    except exceptions.UsageError as e:
        cmdHelp['help'] = str(e)
    for cmd in Commands:
        # print(cmd)
        if cmd == 'help':
            continue
        try:
            cmdenv(['trade', cmd, '-h'])
        except exceptions.UsageError as e:
            cmdHelp[cmd] = str(e)
        index = commands.commandIndex[cmd]
        
        allArgs[cmd] = {'req': {}, 'opt': {}}
        if index.arguments:
            for arg in index.arguments:
                # print(arg.args[0])
                argVals[arg.args[0]] = arg.kwargs.get('default') or None
                
                allArgs[cmd]['req'][arg.args[0]] = {kwarg: arg.kwargs[kwarg] for kwarg in arg.kwargs}
                allArgs[cmd]['req'][arg.args[0]]['widget'] = chooseType(arg)
        # print(allArgs[cmd]['req'])
        
        if index.switches:
            for arg in index.switches:
                try:
                    argVals[arg.args[0]] = arg.kwargs.get('default') or None
                    
                    allArgs[cmd]['opt'][arg.args[0]] = {kwarg: arg.kwargs[kwarg] for kwarg in arg.kwargs}
                    allArgs[cmd]['opt'][arg.args[0]]['widget'] = chooseType(arg)
                    
                    if arg.args[0] == '--option':
                        # Currently only the 'import' cmd has the '--plug' option,
                        # but this could no longer be the case in future.
                        if cmd == 'import':
                            plugOptions = {
                                plug: plugins.load(cmdenv(['trade', cmd, '--plug', plug, '-O', 'help']).plug,
                                                    "ImportPlugin").pluginOptions for plug in importPlugs
                                }
                            allArgs[cmd]['opt'][arg.args[0]]['options'] = plugOptions
                
                except AttributeError:
                    for argGrp in arg.arguments:
                        argVals[argGrp.args[0]] = argGrp.kwargs.get('default') or None
                        
                        allArgs[cmd]['opt'][argGrp.args[0]] = {kwarg: argGrp.kwargs[kwarg] for kwarg in argGrp.kwargs}
                        allArgs[cmd]['opt'][argGrp.args[0]]['widget'] = chooseType(argGrp)
                        
                        allArgs[cmd]['opt'][argGrp.args[0]]['excludes'] = [excl.args[0] for excl in arg.arguments
                                                                   if excl.args[0] != argGrp.args[0]]
                        if argGrp.args[0] == '--plug':
                            # Currently only the 'import' cmd has the '--plug' option,
                            # but this could no longer be the case in the future.
                            if cmd == 'import':
                                allArgs[cmd]['opt'][argGrp.args[0]]['plugins'] = importPlugs


def optWindow():
    """
    Opens a window listing all of the options for the currently selected plugin.
    """
    # Create a new subwindow for plugin options
    sw = TDWidget.SubWindow("Plugin Options", modal=True)
    
    # Clear any previous widgets
    sw.clear()
    
    # Build dictionary of current option values
    optDict = {}
    currentOptStr = argVals.get('--option')
    if currentOptStr:
        for option in currentOptStr.split(','):
            if '=' in option:
                key, val = option.split('=', 1)
                optDict[key] = val
            elif option != '':
                optDict[option] = True
    
    # Check if a plugin is selected
    selectedPlug = sw.combo('--plug')
    if not selectedPlug:
        sw.message('No import plugin chosen.', width=170)
    else:
        plugOpts = allArgs['import']['opt']['--option']['options'].get(selectedPlug, {})
        for option, tooltip in plugOpts.items():
            if '=' in tooltip:
                sw.entry(option, value=optDict.get(option, ''), label=True, sticky='ew', tooltip=tooltip)
            else:
                sw.check(option, optDict.get(option, False), sticky='ew', tooltip=tooltip)
    
    # Add Done and Cancel buttons
    sw.button("Done", setOpts)
    sw.button("Cancel", sw.hide)
    
    # Display the subwindow
    sw.show()


def chooseType(arg):
    """
    Choose what type of TDWidget to make for the passed argument
    """
    if arg.kwargs.get('action') in ('store_true', 'store_const'):
        return {'type': 'check'}
    if arg.kwargs.get('type') == int:
        return {'type': 'spin', 'min': 0, 'max': 4096}
    if arg.kwargs.get('choices'):
        return {'type': 'ticks', 'values': list(arg.kwargs.get('choices'))}
    if arg.args[0] == '--plug':
        return {'type': 'combo', 'values': [''] + importPlugs}
    if arg.args[0] == '--option':
        return {'type': 'option', 'func': optWindow}
    if arg.kwargs.get('type') == float:
        return {'type': 'numeric'}
    if arg.kwargs.get('type') == 'credits':
        return {'type': 'credits'}
    return {'type': 'entry'}

def addWidget(widgetType, parent=None, cpos=0, rpos=0, **kwargs):
    """
    Adds a new TDWidget widget and configures it based on passed parameters
    """
    cspan = kwargs.pop('colspan', None)
    rspan = kwargs.pop('rowspan', None)
    
    # Create widget based on type
    if widgetType == 'combo':
        widget = TDWidget.Combo(parent, values=kwargs.get('values'), textvariable=kwargs.get('textvariable'))
    elif widgetType == 'ticks':
        widget = TDWidget.List(parent, values=kwargs.get('values'), selectmode=kwargs.get('selectmode'), height=kwargs.get('height'))
    elif widgetType == 'stext':
        widget = TDWidget.ScrolledText(parent, textvariable=kwargs.get('textvariable'))
    elif widgetType == 'button':
        widget = TDWidget.Button(parent, text=kwargs.get('text'), command=kwargs.get('func'))
    elif widgetType == 'frame':
        widget = TDWidget.Frame(parent)
    elif widgetType == 'tab':
        widget = TDWidget.Tab(parent)
    elif widgetType == 'label':
        widget = TDWidget.Label(parent, text=kwargs.get('text'))
    elif widgetType == 'check':
        widget = TDWidget.Check(parent, text=kwargs.get('text'), variable=kwargs.get('textvariable'),
                                onvalue=kwargs.get('onvalue', True), offvalue=kwargs.get('offvalue', False),
                                command=kwargs.get('func'))
    elif widgetType == 'spin':
        widget = TDWidget.Spin(parent, min=kwargs.get('min', 0), max=kwargs.get('max', 100),
                               textvariable=kwargs.get('textvariable'))
    else:  # default entry
        widget = TDWidget.Entry(parent, textvariable=kwargs.get('textvariable'))
    
    # Set common attributes
    if 'font' in kwargs:
        widget.font = kwargs['font']
    if 'sticky' in kwargs:
        widget.sticky = kwargs['sticky']
    if 'width' in kwargs:
        widget.width = kwargs['width']
    if 'justify' in kwargs:
        widget.justify = kwargs['justify']
    if 'height' in kwargs:
        widget.height = kwargs['height']
    if 'state' in kwargs:
        widget.state = kwargs['state']
    if 'default' in kwargs:
        widget.set(kwargs['default'])
    
    # Place widget in grid
    widget.grid(column=cpos, row=rpos, columnspan=cspan, rowspan=rspan)
    
    # Handle tabs
    if 'tab' in kwargs and hasattr(parent, 'add'):
        parent.add(widget, text=kwargs['tab'])
    
    return widget

def addWidgetFromArg(name, arg, parent):
    """
    Creates a labeled TDWidget for an argument.
    """
    widgets[name] = TDWidget.Frame(parent)
    
    kwargs = arg['widget'].copy()
    kwargs['textvariable'] = argVals[name]
    widgetType = kwargs.pop('type', None)
    
    # Handle special types
    if widgetType == 'ticks':
        kwargs['height'] = len(kwargs['values'])
        argVals[name] = value = kwargs.pop('values', None)
        kwargs['listvariable'] = argVals[name]
        kwargs['selectmode'] = 'extended'
    elif widgetType == 'numeric':
        pass
    elif widgetType == 'credits':
        pass
    
    # Create label or button for non-check widgets
    if widgetType == 'check':
        kwargs['text'] = name
        kwargs['columnspan'] = 3
    else:
        if widgetType == 'option':
            label = TDWidget.Button(widgets[name], text=name, command=kwargs.pop('func', None))
            widgetType = 'entry'
        else:
            label = TDWidget.Label(widgets[name], text=name)
        label.grid(column=0, row=0)
        kwargs['rpos'] = 0
        kwargs['cpos'] = 1
        kwargs['columnspan'] = 2
    
    # Add the actual input widget
    addWidget(widgetType, parent=widgets[name], **kwargs)
    
    # Place the container frame
    widgets[name].grid()
    
    def makeWidgets(name, arg, sticky='ew', label=True, **kwargs):
        """
        Creates and places a TDWidget for the given argument, handling all types.
        """
        kwargs['sticky'] = sticky
        kwargs['label'] = label
        kwargs['change'] = updArgs
        kwargs['tooltip'] = arg.get('help', '')
        kwargs['colspan'] = 1 if arg == allArgs.get(name) else 9
        
        widgetDef = arg['widget']
        
        # Button
        if widgetDef['type'] == 'button':
            kwargs.pop('change', None)
            kwargs.pop('label', None)
            kwargs.pop('colspan', None)
            TDWidget.Button(win, text=name, command=widgetDef.get('func'), **kwargs)
        
        # Checkbutton
        elif widgetDef['type'] == 'check':
            TDWidget.Check(win, text=name, variable=argVals[name] or arg.get('default'), **kwargs)
        
        # Spinbox
        elif widgetDef['type'] == 'spin':
            kwargs['item'] = argVals[name] or arg.get('default') or 0
            TDWidget.Spin(win, from_=-widgetDef.get('range', 0), to=widgetDef.get('range', 0), **kwargs)
        
        # Combobox / ticks
        elif widgetDef['type'] == 'combo':
            kwargs['sticky'] = 'w'
            if widgetDef.get('sub'):
                kwargs['kind'] = widgetDef['sub']
                kwargs.pop('label', None)
            
            combo = TDWidget.Combo(win, values=widgetDef.get('values', []), variable=argVals[name], **kwargs)
            
            if not widgetDef.get('sub'):
                argVals[name] = argVals[name] or arg.get('default') or '?'
                combo.set(argVals[name])
            else:
                if argVals[name]:
                    for val, vval in argVals[name].items():
                        combo.setOption(val, values=vval, callFunction=False)
        
        # Option button + entry
        elif widgetDef['type'] == 'option':
            kwargs.pop('change', None)
            kwargs.pop('label', None)
            kwargs.pop('colspan', None)
            TDWidget.Button(win, text='optionButton', command=optionsWin, name='--option', **kwargs)
            
            kwargs['sticky'] = sticky
            kwargs['change'] = updArgs
            TDWidget.Entry(win, textvariable=argVals[name] or arg.get('default'), row='p', column=1, colspan=9, **kwargs)
        
        # Entry / numeric / credits
        elif widgetDef['type'] == 'entry':
            if widgetDef.get('sub') == 'credits':
                # TODO: Implement credits type handling
                pass
            elif widgetDef.get('sub'):
                kwargs['kind'] = 'numeric'
            
            TDWidget.Entry(win, textvariable=argVals[name] or arg.get('default'), **kwargs)


def updateCommandBox(args=None):
    """
    Updates the argument panes when the selected command is changed using TDWidget.
    """
    cmd = widgets['Command'].get()
    
    # Update help pane
    helpPane = widgets['helpPane']
    helpPane.config(state='normal')
    helpPane.delete(0.0, 'end')
    helpPane.insert(0.0, cmdHelp.get(cmd, ''))
    helpPane.config(state='disabled')
    
    # Hide all currently displayed widgets in 'req' and 'opt'
    for child in widgets['req'].winfo_children():
        child.grid_forget()
    for child in widgets['opt'].winfo_children():
        child.grid_forget()
    
    if cmd == 'help':
        return
    
    # Prepend for 'station' command to avoid name conflicts
    prepend = f"{cmd}~" if cmd == 'station' else ''
    
    # Handle required arguments
    if allArgs[cmd]['req']:
        if 'Required:' not in widgets:
            widgets['Required:'] = TDWidget.Label(widgets['req'], text='Required:', sticky='nw')
        else:
            widgets['Required:'].grid()
        
        for i, key in enumerate(allArgs[cmd]['req'], start=1):
            fullKey = prepend + key
            if fullKey not in widgets:
                addWidgetFromArg(fullKey, allArgs[cmd]['req'][key], widgets['req'])
            else:
                widgets[fullKey].grid(column=0, row=i)
    
    # Handle optional arguments
    if allArgs[cmd]['opt']:
        if 'Optional:' not in widgets:
            widgets['Optional:'] = TDWidget.Label(widgets['opt'], text='Optional:', sticky='nw')
        else:
            widgets['Optional:'].grid()
        
        for i, key in enumerate(allArgs[cmd]['opt'], start=1):
            fullKey = prepend + key
            if fullKey not in widgets:
                addWidgetFromArg(fullKey, allArgs[cmd]['opt'][key], widgets['opt'])




# Setup the CLI interface and build the main window
def main(argv=None):
    class IORedirector:
        def __init__(self, TEXT_INFO):
            self.TEXT_INFO = TEXT_INFO
    
    class StdoutRedirector(IORedirector):
        def write(self, string):
            current = self.TEXT_INFO.cget('text').rsplit('\r', 1)[0]
            self.TEXT_INFO.config(text=current + string)
        
        def flush(self):
            sys.__stdout__.flush()
    
    def setOpts():
        """
        Sets the main window options entry to the checked values in the options window.
        """
        sw = widgets.get("Plugin Options")  # TDWidget window reference
        plug = argVals.get('--plug')
        plugOpts = allArgs['import']['opt']['--option']['options'].get(plug, {})
        argStr = ''
        
        if plugOpts:
            for option, val in plugOpts.items():
                w = sw.get(option)  # Retrieve the TDWidget by name
                if '=' in val:
                    if w and isinstance(w, TDWidget.Entry):
                        argStr += f"{option}={w.get()},"  # get() for TDWidget entry
                else:
                    if w and isinstance(w, TDWidget.Checkbutton) and w.get():
                        argStr += f"{option},"
            
            argStr = argStr.rstrip(',')
            if '--option' in widgets:
                widgets['--option'].set(argStr)  # Update main window entry
        
        if sw:
            sw.hide()
    
    def optionsWin():
        """
        Opens a window listing all of the options for the currently selected plugin.
        """
        # Create a new TDWidget window
        sw = TDWidget.SubWindow(title="Plugin Options", modal=True)
        widgets["Plugin Options"] = sw  # store reference for later access
        sw.clear()  # empty the container
        
        optDict = {}
        current_opts = argVals.get('--option')
        if current_opts:
            for option in current_opts.split(','):
                if '=' in option:
                    key, val = option.split('=')
                    optDict[key] = val
                elif option != '':
                    optDict[option] = True
        
        plug = argVals.get('--plug')
        if not plug:
            TDWidget.Label(sw, text="No import plugin chosen.", sticky='ew', colspan=10)
        else:
            plugOpts = allArgs['import']['opt']['--option']['options'].get(plug, {})
            for option, tooltip in plugOpts.items():
                if '=' in tooltip:
                    TDWidget.Entry(
                        sw,
                        name=option,
                        textvariable=optDict.get(option, ''),
                        label=True,
                        sticky='ew',
                        colspan=10,
                        tooltip=tooltip
                    )
                else:
                    TDWidget.Checkbutton(
                        sw,
                        name=option,
                        textvariable=optDict.get(option, False),
                        sticky='ew',
                        colspan=10,
                        tooltip=tooltip
                    )
        
        # Buttons
        TDWidget.Button(sw, text="Done", func=setOpts, column=8)
        TDWidget.Button(sw, text="Cancel", func=sw.hide, row='p', column=9)
        
        sw.show()
    
    def updArgs(name):
        """
        Updates the value of argVals[name] when the linked widget's value is changed in the window.
        """
        
        def getWidget(name):
            """
            Returns the TDWidget instance for the given argument name.
            """
            return widgets.get(name)
        
        # Clear dependent options if the plugin changed
        if name == '--plug' and argVals.get(name) != (w := getWidget(name)).get():
            option_widget = widgets.get('--option')
            if option_widget:
                option_widget.set('')
        
        # Update the stored value
        w = getWidget(name)
        if w:
            argVals[name] = w.get()
        else:
            argVals[name] = None
        
        # Determine excluded arguments for this argument
        if allArgs.get(name):
            excluded = allArgs[name].get('excludes', [])
            argBase = allArgs
        else:
            try:
                cmd = widgets['Command'].get()
                excluded = allArgs[cmd]['opt'][name].get('excludes', [])
                argBase = allArgs[cmd]['opt']
            except KeyError:
                excluded = []
        
        # Reset any arguments excluded by this one if it has a value
        if excluded and argVals.get(name):
            for exclude in excluded:
                w = widgets.get(exclude)
                if not w:
                    continue
                widgetType = argBase[exclude]['widget']['type']
    
    def updCmd():
        """
        Updates the argument panes when the selected command is changed.
        """
        cmd = widgets['Command'].get()
        
        # Update help text
        help_widget = widgets.get('helpPane')
        if help_widget:
            help_widget.set(cmdHelp.get(cmd, ''))
        
        # Clear current argument frames
        for child in widgets['req'].winfo_children():
            child.grid_forget()
        for child in widgets['opt'].winfo_children():
            child.grid_forget()
        
        # Nothing more to do for 'help' command
        if cmd == 'help':
            return
        
        # Show required arguments
        req_frame = widgets['req']
        if allArgs[cmd]['req']:
            if 'Required:' not in widgets:
                widgets['Required:'] = TDWidget.Label(req_frame, text='Required:', sticky='w')
                widgets['Required:'].grid(column=0, row=0, sticky='w')
            else:
                widgets['Required:'].grid()
            
            i = 1
            for key, arg in allArgs[cmd]['req'].items():
                if key not in widgets:
                    addWidgetFromArg(key, arg, req_frame)
                else:
                    widgets[key].grid(column=0, row=i)
                i += 1
        
        # Show optional arguments
        opt_frame = _
    
    def runTD():
        """
        Executes the TD command selected in the GUI using TDWidget.
        """
        
        from . import tradeexcept
        
        def getVals(arg, argBase):
            curArg = argVals.get(arg)
            vals = []
            
            # Skip default or empty values
            if curArg is not None and curArg != argBase[arg].get('default'):
                if arg not in ['--detail', '--debug', '--quiet']:
                    vals.append(str(arg))
                
                widget_info = argBase[arg]['widget']
                wtype = widget_info.get('type')
                sub = widget_info.get('sub')
                
                if wtype == 'check':
                    # Checkbuttons already handled by presence in argVals
                    return vals
                
                if sub == 'ticks':
                    choices = ''
                    for choice, selected in curArg.items():
                        if selected:
                            choices += choice
                    vals.append(choices)
                    return vals
                
                if sub == 'credits':
                    # TODO: Handle 'credits' type
                    pass
                
                # Default behavior: append the current value
                vals.append(str(curArg))
            
            if not vals:
                return None
            
            return vals
        
        def runTrade():
            # Disable the Run button while executing
            run_btn = widgets.get('Run')
            if run_btn:
                run_btn.disable()
            
            # Redirect stdout to the output Text widget
            output_widget = widgets.get('outputText')
            oldout = sys.stdout
            sys.stdout = StdoutRedirector(output_widget)
            print('TD command: "' + ' '.join(argv) + '"')
            
            try:
                try:
                    try:
                        if "CPROF" in os.environ:
                            import cProfile
                            cProfile.run("trade(argv)")
                        else:
                            trade(argv)
                    except PluginException as e:
                        print(f"PLUGIN ERROR: {e}")
                        if 'EXCEPTIONS' in os.environ:
                            raise e
                        sys.exit(1)
                    except tradeexcept.TradeException as e:
                        print(f"{argv[0]}: {e}")
                        if 'EXCEPTIONS' in os.environ:
                            raise e
                        sys.exit(1)
                except (UnicodeEncodeError, UnicodeDecodeError):
                    print("-----------------------------------------------------------")
                    print("ERROR: Unexpected unicode error in the wild!")
                    print(traceback.format_exc())
                    print(
                        "Please report this bug (http://kfs.org/td/issues). "
                        "You may be able to work around it using '-q'. "
                        "Windows users can try 'chcp.com 65001' to enable UTF-8."
                    )
            except SystemExit as e:
                print(e)
            
            print("Execution complete.")
            # Restore stdout
            sys.stdout = oldout
            # Re-enable Run button
            if run_btn:
                run_btn.enable()
        
        # Switch to the Output tab
        tab_frame = widgets.get('tabFrame')
        if tab_frame:
            tab_frame.select_tab('Output')
        
        # Build the argv list
        cmd = widgets['Command'].get()
        argv = ['trade', cmd]
        
        if cmd != 'help':
            # Required args
            for arg in allArgs[cmd]['req']:
                result = getVals(arg, allArgs[cmd]['req'])
                if result:
                    argv += result if '-' in result[0] else result[1:]
            
            # Optional args
            for arg in allArgs[cmd]['opt']:
                result = getVals(arg, allArgs[cmd]['opt'])
                if result:
                    argv += result
            
            # Global args
            for arg in allArgs:
                if arg in Commands:
                    continue
                result = getVals(arg, allArgs)
                if result:
                    argv += result
        
        # Clear previous output
        if output_widget:
            output_widget.set('')
        
        # Run trade in a separate thread
        threading.Thread(target=runTrade, name="TDThread", daemon=True).start()
    
    def makeWidgets(name, arg, sticky='ew', label=True, **kwargs):
        """
        Creates and places a widget for the given argument using TDWidget.
        """
        kwargs['sticky'] = sticky
        kwargs['label'] = label
        kwargs['change'] = updArgs
        kwargs['tooltip'] = arg.get('help', '')
        kwargs['colspan'] = 1 if arg == allArgs.get(name) else 9
        
        widget_info = arg['widget']
        wtype = widget_info['type']
        
        if wtype == 'button':
            kwargs.pop('change', None)
            kwargs.pop('label', None)
            kwargs.pop('colspan', None)
            TDWidget(
                name=name,
                widget_type='button',
                command=widget_info.get('func'),
                **kwargs
            )
        
        elif wtype == 'check':
            TDWidget(
                name=name,
                widget_type='check',
                values=argVals.get(name, arg.get('default', False)),
                text=name,
                **kwargs
            )
        
        elif wtype == 'spin':
            TDWidget(
                name=name,
                widget_type='spin',
                values=argVals.get(name, arg.get('default', 0) or 0),
                min_value=widget_info.get('min', 0),
                max_value=widget_info.get('max', 100),
                **kwargs
            )
        
        elif wtype == 'combo':
            kwargs['sticky'] = 'w'
            if widget_info.get('sub'):
                kwargs['kind'] = widget_info['sub']
                kwargs.pop('label', None)
            
            combo = TDWidget(
                name=name,
                widget_type='combo',
                values=widget_info.get('values', []),
                **kwargs
            )
            
            if not widget_info.get('sub'):
                if not isinstance(argVals.get(name), str):
                    argVals[name] = None
                default_val = '?' if arg.get('choices') else ''
                combo.set(argVals.get(name) or arg.get('default') or default_val, callFunction=False)
            else:
                if isinstance(argVals.get(name), str):
                    argVals[name] = None
                if argVals.get(name):
                    for val, vval in argVals[name].items():
                        combo.set_option(val, values=vval, callFunction=False)
        
        elif wtype == 'option':
            kwargs.pop('change', None)
            kwargs.pop('label', None)
            kwargs.pop('colspan', None)
            
            # Option button opens the plugin options window
            TDWidget(
                name='optionButton',
                widget_type='button',
                command=optionsWin,
                option_name='--option',
                **kwargs
            )
            
            kwargs['sticky'] = sticky
            kwargs['change'] = updArgs
            TDWidget(
                name=name,
                widget_type='entry',
                values=argVals.get(name) or arg.get('default', ''),
                row='p',
                column=1,
                colspan=9,
                **kwargs
            )
        
        elif wtype == 'entry':
            if widget_info.get('sub') == 'credits':
                # TODO: Handle 'credits' type
                pass
            elif widget_info.get('sub'):
                kwargs['kind'] = 'numeric'
            
            TDWidget(
                name=name,
                widget_type='entry',
                values=argVals.get(name) or arg.get('default', ''),
                **kwargs
            )
    
    sys.argv = ['trade']
    if not argv:
        argv = sys.argv
    if sys.hexversion < 0x30813F0:
        raise SystemExit(
            "Sorry: TradeDangerous requires Python 3.8.19 or higher.\n"
            "For assistance, see:\n"
            "\tBug Tracker: https://github.com/eyeonus/Trade-Dangerous/issues\n"
            "\tDocumentation: https://github.com/eyeonus/Trade-Dangerous/wiki\n"
            "\tEDForum Thread: https://forums.frontier.co.uk/showthread.php/441509\n"
            )
    
    
    buildArgDicts()
    
    # --- Root window ---
    main_win = TDWidget(name='root', widget_type='frame')
    main_win.widget.master.title('Trade Dangerous GUI (Beta), TD v.%s' % (__version__,))
    
    # --- Command Combo ---
    widgets['Command'] = TDWidget(
        name='Command', widget_type='combo', parent=main_win,
        values=Commands, width=10, row=0, column=0, columnspan=5, sticky='ew',
        command=updCmd
    )
    
    # --- Request / Optional Scroll Frames ---
    widgets['req'] = TDWidget('req', 'frame', parent=main_win, row=1, column=0, columnspan=10, sticky='nsew')
    widgets['req'].widget.config(width=200, height=75)
    widgets['opt'] = TDWidget('opt', 'frame', parent=main_win, row=2, column=0, columnspan=10, sticky='nsew')
    widgets['opt'].widget.config(width=200, height=345)
    
    # --- Tabbed Frame ---
    tabFrame = TDWidget('tabFrame', 'notebook', parent=main_win, row=1, column=10, rowspan=2, columnspan=40, sticky='nsew')
    
    # Help tab
    help_tab = TDWidget('helpTab', 'frame', parent=tabFrame)
    tabFrame.widget.add(help_tab.widget, text='Help')
    widgets['helpPane'] = TDWidget('helpPane', 'scrolledtext', parent=help_tab, row=0, column=0, width=80, height=25)
    widgets['helpPane'].set(cmdHelp['help'])
    widgets['helpPane'].widget.config(state='disabled', width=80)
    
    # Output tab
    output_tab = TDWidget('outputTab', 'frame', parent=tabFrame)
    tabFrame.widget.add(output_tab.widget, text='Output')
    widgets['outPane'] = TDWidget('outPane', 'scrolledtext', parent=output_tab, row=0, column=0, width=80, height=25)
    widgets['outPane'].set('')
    widgets['outPane'].widget.config(state='disabled', width=80)
    
    # --- Option Widgets ---
    makeWidgets('--link-ly', allArgs['--link-ly'], sticky='w', width=4, row=3, column=2)
    makeWidgets('--quiet', allArgs['--quiet'], sticky='e', disabled=':', width=1, row=3, column=46)
    makeWidgets('--detail', allArgs['--detail'], sticky='e', disabled=':', width=1, row=3, column=47)
    makeWidgets('--debug', allArgs['--debug'], sticky='e', disabled=':', width=1, row=3, column=48)
    
    # --- Run Button ---
    TDWidget('Run', 'button', parent=main_win, text='Run', command=runTD, row=3, column=49, sticky='w')
    
    # --- CWD ---
    makeWidgets('--cwd', allArgs['--cwd'], width=4, row=4, column=0)
    cwd_scroll = TDWidget('CWD', 'scrolledtext', parent=main_win, row=4, column=1, columnspan=49, width=70, height=1)
    cwd_scroll.set(argVals['--cwd'])
    cwd_scroll.widget.config(state='disabled')
    widgets['cwd'] = TDWidget('cwd', 'label', parent=main_win, text=argVals['--cwd'], sticky='w', row=4, column=1)
    
    # --- DB ---
    makeWidgets('--db', allArgs['--db'], width=4, row=5, column=0)
    db_scroll = TDWidget('DB', 'scrolledtext', parent=main_win, row=5, column=1, columnspan=49, width=70, height=1)
    db_scroll.set(argVals['--db'])
    db_scroll.widget.config(state='disabled')
    widgets['db'] = TDWidget('db', 'label', parent=main_win, text=argVals['--db'], sticky='w', row=5, column=1)
    
    # --- Configure row/column stretching for proper layout ---
    for i in range(50):
        main_win.widget.columnconfigure(i, weight=1)
    for i in range(6):
        main_win.widget.rowconfigure(i, weight=1)
    
    # --- Show window ---
    main_win.widget.mainloop()
    
    # with gui('Trade Dangerous GUI (Beta), TD v.%s' % (__version__,), inPadding=1) as win:
    #     win.setFont(size=8, family='Courier')
    #     win.combo('Command', Commands, change=updCmd, tooltip='Trade Dangerous command to run.',
    #               stretch='none', sticky='ew', width=10, row=0, column=0, colspan=5)
    #     with win.scrollPane('req', disabled='horizontal', row=1, column=0, colspan=10) as pane:
    #         pane.configure(width=200, height=75)
    # 
    #     with win.scrollPane('opt', disabled='horizontal', row=2, column=0, colspan=10) as pane:
    #         pane.configure(width=200, height=345)
    # 
    #     with win.tabbedFrame('tabFrame', disabled='horizontal', row=1, column=10, rowspan=2, colspan=40) as tabFrame:
    #         with win.tab('Help'):
    #             with win.scrollPane('helpPane', disabled='horizontal') as pane:
    #                 pane.configure(width=560, height=420)
    #                 win.message('helpText', cmdHelp['help'])
    #                 win.widgetManager.get(WIDGET_NAMES.Message, 'helpText').config(width=560)
    # 
    #         with win.tab('Output'):
    #             with win.scrollPane('outPane', disabled='horizontal') as pane:
    #                 pane.configure(width=560, height=420)
    #                 win.message('outputText', '')
    #                 win.widgetManager.get(WIDGET_NAMES.Message, 'outputText').config(width=560)
    # 
    #     makeWidgets('--link-ly', allArgs['--link-ly'], sticky='w', width=4, row=3, column=2)
    # 
    #     makeWidgets('--quiet', allArgs['--quiet'], sticky='e', disabled=':', width=1, row=3, column=46)
    # 
    #     makeWidgets('--detail', allArgs['--detail'], sticky='e', disabled=':', width=1, row=3, column=47)
    # 
    #     makeWidgets('--debug', allArgs['--debug'], sticky='e', disabled=':', width=1, row=3, column=48)
    # 
    #     win.button('Run', runTD, tooltip='Execute the selected command.',
    #                sticky='w', row=3, column=49)
    # 
    #     makeWidgets('--cwd', allArgs['--cwd'], width=4, row=4, column=0)
    #     with win.scrollPane('CWD', disabled='vertical', row=4, column=1, colspan=49) as pane:
    #         pane.configure(width=500, height=20)
    #         widgets['cwd'] = win.label('cwd', argVals['--cwd'], sticky='w')
    # 
    #     makeWidgets('--db', allArgs['--db'], width=4, row=5, column=0)
    #     with win.scrollPane('DB', disabled='vertical', row=5, column=1, colspan=49) as pane:
    #         pane.configure(width=500, height=20)
    #         widgets['db'] = win.label('db', argVals['--db'], sticky='w')


def trade(argv):
    """
    This method represents the trade command.
    """
    cmdIndex = commands.CommandIndex()
    cmdenv = cmdIndex.parse(argv)
    
    tdb = tradedb.TradeDB(cmdenv, load = cmdenv.wantsTradeDB)
    if cmdenv.usesTradeData:
        tsc = tdb.tradingStationCount
        if tsc == 0:
            raise exceptions.NoDataError(
                "There is no trading data for ANY station in "
                "the local database. Please enter or import "
                "price data."
            )
        if tsc == 1:
            raise exceptions.NoDataError(
                "The local database only contains trading data "
                "for one station. Please enter or import data "
                "for additional stations."
            )
        if tsc < 8:
            cmdenv.NOTE(
                "The local database only contains trading data "
                "for {} stations. Please enter or import data "
                "for additional stations.".format(
                    tsc
                )
            )
    
    try:
        results = cmdenv.run(tdb)
    finally:
        # always close tdb
        tdb.close()
    
    if results:
        results.render()



# tradedangerous/db/engine.py
from __future__ import annotations
from pathlib import Path
from typing import Any, Dict, Mapping
import configparser
import os
import time

from sqlalchemy import create_engine, event, text
from sqlalchemy.engine import Engine, URL
from sqlalchemy.orm import sessionmaker, Session  # type: ignore
from sqlalchemy.pool import NullPool
from sqlalchemy.exc import OperationalError

from .config import DEFAULTS, load_config
from .paths import resolve_data_dir, resolve_tmp_dir, resolve_db_config_path


# ---------- config normalization & helpers ----------

def _ensure_default_config_file(target_path: Path | None) -> Path | None:
    """
    If *target_path* is provided and no file exists there, write a minimal db_config.ini
    built from in-code DEFAULTS. Returns the path if created, else None.
    """
    if not target_path:
        return None
    if target_path.exists():
        return target_path
    # Build from DEFAULTS
    target_path.parent.mkdir(parents=True, exist_ok=True)
    cp = configparser.ConfigParser()
    for section, mapping in DEFAULTS.items():
        cp[section] = {}
        if isinstance(mapping, Mapping):
            for k, v in mapping.items():
                cp[section][k] = str(v)
    with target_path.open("w", encoding="utf-8") as fh:
        cp.write(fh)
    return target_path


def _cfg_to_dict(cfg: configparser.ConfigParser | Mapping[str, Any] | str | os.PathLike) -> Dict[str, Dict[str, Any]]:
    """
    Normalise configuration input into a dict-of-sections.
    
    Accepted inputs:
      * dict-like mapping → returned as {section: {key: value}}
      * ConfigParser       → converted to nested dict (sections overlay DEFAULT section)
      * str/Path           → if file exists, read it; if missing, fall back to load_config()
    
    NOTE:
    - We do NOT raise on a missing path; we delegate to load_config() to honour the
      documented resolution order (ENV → CWD → DEFAULTS).
    """
    if isinstance(cfg, (str, os.PathLike)):
        p = Path(cfg)
        if p.exists():
            cp = configparser.ConfigParser()
            with p.open("r", encoding="utf-8") as fh:
                cp.read_file(fh)
            return _cfg_to_dict(cp)
        # Missing provided path → use canonical loader with fallbacks
        return load_config(None)
    
    if isinstance(cfg, configparser.ConfigParser):
        out: Dict[str, Dict[str, Any]] = {}
        defaults = dict(cfg.defaults())
        for sec in cfg.sections():
            d = dict(defaults)
            d.update({k: v for k, v in cfg.items(sec)})
            out[sec] = d
        for sec in ("database", "engine", "sqlite", "mariadb", "paths"):
            out.setdefault(sec, dict(defaults))
        return out
    
    # Already a dict-like mapping of sections
    return {k: dict(v) if isinstance(v, Mapping) else {} for k, v in cfg.items()}  # type: ignore[arg-type]


def _get(cfg: Dict[str, Any], section: str, key: str, default=None):
    if section in cfg and key in cfg[section]:
        return cfg[section][key]
    if "database" in cfg and key in cfg["database"]:
        return cfg["database"][key]
    return default


def _get_int(cfg: Dict[str, Any], section: str, key: str, default=None):
    try:
        return int(_get(cfg, section, key, default))
    except (TypeError, ValueError):
        return default


def _get_bool(cfg: Dict[str, Any], section: str, key: str, default=None):
    v = _get(cfg, section, key, default)
    if isinstance(v, bool):
        return v
    if isinstance(v, str):
        return v.strip().lower() in {"1", "true", "yes", "on"}
    return default


# ---------- URL builders ----------

def _redact(url: str) -> str:
    if "://" not in url:
        return url
    head, rest = url.split("://", 1)
    if "@" in rest and ":" in rest.split("@", 1)[0]:
        user_pass, host = rest.split("@", 1)
        user = user_pass.split(":", 1)[0]
        return f"{head}://{user}:***@{host}"
    return f"{head}://{rest}"

def _make_mariadb_url(cfg: Dict[str, Any]) -> URL:
    driver = str(_get(cfg, "mariadb", "driver", "mariadbconnector")).strip().lower()
    drivername = "mariadb+" + driver if driver == "mariadbconnector" else "mysql+" + driver
    return URL.create(
        drivername=drivername,
        username=str(_get(cfg, "mariadb", "user", "")),
        password=str(_get(cfg, "mariadb", "password", "")),
        host=str(_get(cfg, "mariadb", "host", "127.0.0.1")),
        port=int(_get(cfg, "mariadb", "port", 3306)),
        database=str(_get(cfg, "mariadb", "name", "tradedangerous")),
        query={"charset": str(_get(cfg, "mariadb", "charset", "utf8mb4"))},
    )


def _make_sqlite_url(cfg: Dict[str, Any]) -> str:
    data_dir = resolve_data_dir(cfg)
    # Honour legacy filename
    filename = str(_get(cfg, "sqlite", "sqlite_filename", "TradeDangerous.db"))
    db_path = (data_dir / filename).resolve()
    return f"sqlite+pysqlite:///{db_path.as_posix()}"


# ---------- Engine construction ----------

def make_engine_from_config(cfg_or_path: configparser.ConfigParser | Mapping[str, Any] | str | os.PathLike | None = None) -> Engine:
    """
    Build a SQLAlchemy Engine for either MariaDB or SQLite.
    
    Accepts: ConfigParser, dict-like {section:{k:v}}, path to INI file, or None.
    First-run behaviour:
      - If a path is provided but missing, or if no path is provided and no config is found,
        a default db_config.ini is CREATED in the resolved default location (CWD unless TD_DB_CONFIG
        points elsewhere), then loaded.
    """
    ini_target: Path | None = None
    
    # If caller gave a specific path, prefer to materialise a default file there.
    if isinstance(cfg_or_path, (str, os.PathLike)):
        ini_target = Path(cfg_or_path)
        _ensure_default_config_file(ini_target)
    else:
        # No specific path: create (if missing) at the standard location
        # (CWD/db_config.ini by default, or the file pointed to by TD_DB_CONFIG).
        ini_target = resolve_db_config_path("db_config.ini")
        _ensure_default_config_file(ini_target)
    
    cfg = _cfg_to_dict(cfg_or_path if cfg_or_path is not None else str(ini_target))
    
    # Ensure dirs exist (used by various parts of the app)
    _ = resolve_data_dir(cfg)
    _ = resolve_tmp_dir(cfg)
    
    backend = str(_get(cfg, "database", "backend", "sqlite")).strip().lower()
    echo = bool(_get_bool(cfg, "engine", "echo", False))
    isolation = _get(cfg, "engine", "isolation_level", None)
    
    if backend == "mariadb":
        url = _make_mariadb_url(cfg)
        connect_timeout = _get_int(cfg, "engine", "connect_timeout", 10) or 10
        pool_size    = _get_int(cfg, "engine", "pool_size", 10) or 10
        max_overflow = _get_int(cfg, "engine", "max_overflow", 20) or 20
        pool_timeout = _get_int(cfg, "engine", "pool_timeout", 30) or 30
        pool_recycle = _get_int(cfg, "engine", "pool_recycle", 1800) or 1800
        engine = create_engine(
            url,
            echo=echo,
            pool_pre_ping=True,
            pool_size=pool_size,
            max_overflow=max_overflow,
            pool_timeout=pool_timeout,
            pool_recycle=pool_recycle,
            isolation_level=isolation or "READ COMMITTED",
            connect_args={"connect_timeout": connect_timeout},
        )
    elif backend == "sqlite":
        url = _make_sqlite_url(cfg)
        engine = create_engine(
            url,
            echo=echo,
            poolclass=NullPool,
            connect_args={"check_same_thread": False},
        )
        
        @event.listens_for(engine, "connect")
        def _set_sqlite_pragmas(dbapi_conn, _):
            cur = dbapi_conn.cursor()
            cur.execute("PRAGMA foreign_keys=ON")
            cur.execute("PRAGMA synchronous=OFF")
            cur.execute("PRAGMA temp_store=MEMORY")
            cur.execute("PRAGMA auto_vacuum=INCREMENTAL")
            cur.close()
    else:
        raise ValueError(f"Unsupported backend: {backend}")
    
    try:
        engine._td_redacted_url = _redact(str(url))  # type: ignore[attr-defined]
    except Exception:
        pass
    return engine


# ---------- Session factory ----------

def get_session_factory(engine: Engine) -> sessionmaker[Session]:
    return sessionmaker(bind=engine, expire_on_commit=False, autoflush=True)


# ---------- Health helpers ----------

def healthcheck(engine: Engine, retries: int = 0) -> bool:
    attempt = 0
    delay = 0.25
    while True:
        try:
            with engine.connect() as conn:
                conn.execute(text("SELECT 1"))
            return True
        except OperationalError:
            attempt += 1
            if attempt > retries:
                return False
            time.sleep(delay)
            delay *= 2


def read_sqlite_pragmas(engine: Engine) -> Dict[str, Any]:
    """
    Return active PRAGMA values (SQLite only). Safe no-op for non-sqlite engines.
    """
    out: Dict[str, Any] = {}
    with engine.connect() as conn:
        if conn.dialect.name != "sqlite":
            return out
        
        def one(q: str) -> Any:
            return conn.execute(text(q)).scalar()
        
        out["foreign_keys"] = one("PRAGMA foreign_keys")
        out["synchronous"]  = one("PRAGMA synchronous")
        out["temp_store"]   = one("PRAGMA temp_store")
        out["auto_vacuum"]  = one("PRAGMA auto_vacuum")
    
    return out
